
from .cauchy_spectral_mixture import CauchyMixtureKernel
from .adaptive_kernel import AdaptiveKernel
from .sinc_kernel import SincKernel


__all__ = [
    "CauchyMixtureKernel",
    "AdaptiveKernel",
    "SincKernel",
]



import logging

import numpy as np
import torch
from gpytorch.constraints.constraints import Interval
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.means import ConstantMean
from gpytorch.mlls import ExactMarginalLogLikelihood
from gpytorch.models import ExactGP
from collections.abc import Callable
import random
from copy import deepcopy

from SMKBO.cas.kernels import CategoricalOverlap, TransformedCategorical, OrdinalKernel, MixtureKernel


def onehot2ordinal(x, categorical_dims):
    """Convert one-hot representation of strings back to ordinal representation."""
    from itertools import chain
    if x.ndim == 1:
        x = x.reshape(1, -1)
    categorical_dims_flattned = list(chain(*categorical_dims))
    # Select those categorical dimensions only
    x = x[:, categorical_dims_flattned]
    if not isinstance(x, torch.Tensor):
        x = torch.tensor(x)
    res = torch.zeros(x.shape[0], len(categorical_dims), dtype=torch.float32)
    for i, var_group in enumerate(categorical_dims):
        res[:, i] = torch.argmax(x[:, var_group], dim=-1).float()
    return res


def ordinal2onehot(x, n_categories):
    """Convert ordinal to one-hot"""
    res = np.zeros(np.sum(n_categories))
    offset = 0
    for i, cat in enumerate(n_categories):
        res[offset + int(x[i])] = 1
        offset += cat
    return torch.tensor(res)


# GP Model
class GP(ExactGP):
    def __init__(self, train_x, train_y, kern, likelihood,
                 outputscale_constraint,
                 ard_dims, cat_dims=None):
        super(GP, self).__init__(train_x, train_y, likelihood)
        self.dim = train_x.shape[1]
        self.ard_dims = ard_dims
        self.cat_dims = cat_dims
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(kern, outputscale_constraint=outputscale_constraint)

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)  # , cat_dims, int_dims)
        return MultivariateNormal(mean_x, covar_x)


def train_gp(train_x, train_y, use_ard, num_steps, kern='transformed_overlap', hypers={},
             cat_dims=None, cont_dims=None,
             int_constrained_dims=None,
             noise_variance=None,
             cat_configs=None,
             continuous_kern_type = 'smk',
             num_mixtures1=5, num_mixtures2=4,
             **params):
    """Fit a GP model where train_x is in [0, 1]^d and train_y is standardized.
    （train_x, train_y）: pairs of x and y (trained)
    noise_variance: if provided, this value will be used as the noise variance for the GP model. Otherwise, the noise
        variance will be inferred from the model.
    int_constrained_dims: **Of the continuous dimensions**, which ones additionally are constrained to have integer
        values only?
    """
    assert train_x.ndim == 2
    assert train_y.ndim == 1
    assert train_x.shape[0] == train_y.shape[0]

    # Create hyper parameter bounds
    if noise_variance is None:
        noise_variance = 0.005
        noise_constraint = Interval(1e-6, 0.1)
    else:
        if np.abs(noise_variance) < 1e-6:
            noise_variance = 0.05
            noise_constraint = Interval(1e-6, 0.1)
        else:
            noise_constraint = Interval(0.99 * noise_variance, 1.01 * noise_variance)
    if use_ard:
        lengthscale_constraint = Interval(0.01, 0.5)
    else:
        lengthscale_constraint = Interval(0.01, 2.5)  # [0.005, sqrt(dim)]
    # outputscale_constraint = Interval(0.05, 20.0)
    outputscale_constraint = Interval(0.5, 5.)

    # Create models
    likelihood = GaussianLikelihood(noise_constraint=noise_constraint).to(device=train_x.device, dtype=train_y.dtype)
    # train_x = onehot2ordinal(train_x, cat_dims)
    ard_dims = train_x.shape[1] if use_ard else None

    if kern == 'overlap':
        kernel = CategoricalOverlap(lengthscale_constraint=lengthscale_constraint, ard_num_dims=ard_dims, )
    elif kern == 'transformed_overlap':
        kernel = TransformedCategorical(lengthscale_constraint=lengthscale_constraint, ard_num_dims=ard_dims, )
    elif kern == 'ordinal':
        kernel = OrdinalKernel(lengthscale_constraint=lengthscale_constraint, ard_num_dims=ard_dims, config=cat_configs)
    elif kern == 'mixed':
        assert cat_dims is not None and cont_dims is not None, 'cat_dims and cont_dims need to be specified if you wish' \
                                                               'to use the mix kernel'
        kernel = MixtureKernel(cat_dims, cont_dims,
                               categorical_ard=use_ard, continuous_ard=use_ard,
                               integer_dims=int_constrained_dims, continuous_kern_type = continuous_kern_type,
                               num_mixtures1=num_mixtures1, num_mixtures2=num_mixtures2,
                               **params)
    elif kern == 'mixed_overlap':
        kernel = MixtureKernel(cat_dims, cont_dims,
                               categorical_ard=use_ard, continuous_ard=use_ard,
                               categorical_kern_type='overlap',
                               integer_dims=int_constrained_dims, continuous_kern_type = continuous_kern_type,
                               num_mixtures1=num_mixtures1, num_mixtures2=num_mixtures2,
                               **params)
    else:
        raise ValueError('Unknown kernel choice %s' % kern)

    model = GP(
        train_x=train_x,
        train_y=train_y,
        likelihood=likelihood,
        kern=kernel,
        # lengthscale_constraint=lengthscale_constraint,
        outputscale_constraint=outputscale_constraint,
        ard_dims=ard_dims,
    ).to(device=train_x.device, dtype=train_x.dtype)

    # Find optimal model hyperparameters
    model.train()
    likelihood.train()

    # "Loss" for GPs - the marginal log likelihood
    mll = ExactMarginalLogLikelihood(likelihood, model)

    # Initialize model hypers
    if hypers:
        model.load_state_dict(hypers)
    else:
        hypers = {}
        hypers["covar_module.outputscale"] = 1.0
        hypers["covar_module.base_kernel.lengthscale"] = np.sqrt(0.01 * 0.5)
        hypers["likelihood.noise"] = noise_variance if noise_variance is not None else 0.005
        model.initialize(**hypers)

    # Use the adam optimizer
    optimizer = torch.optim.Adam([{"params": model.parameters()}], lr=0.03)

    for _ in range(num_steps):
        optimizer.zero_grad()
        output = model(train_x, )
        loss = -mll(output, train_y).float()
        loss.backward()
        optimizer.step()

    # Switch to eval mode
    model.eval()
    likelihood.eval()

    return model


def to_unit_cube(x, lb, ub):
    """Project to [0, 1]^d from hypercube with bounds lb and ub"""
    assert np.all(lb < ub) and lb.ndim == 1 and ub.ndim == 1 and x.ndim == 2
    xx = (x - lb) / (ub - lb)
    return xx


def from_unit_cube(x, lb, ub):
    """Project from [0, 1]^d to hypercube with bounds lb and ub"""
    assert np.all(lb < ub) and lb.ndim == 1 and ub.ndim == 1 and x.ndim == 2
    xx = x * (ub - lb) + lb
    return xx


def latin_hypercube(n_pts, dim):
    """Basic Latin hypercube implementation with center perturbation."""
    X = np.zeros((n_pts, dim))
    centers = (1.0 + 2.0 * np.arange(0.0, n_pts)) / float(2 * n_pts)
    # random.seed(random.randint(0, 1e6))
    for i in range(dim):  # Shuffle the center locataions for each dimension.
        X[:, i] = centers[np.random.permutation(n_pts)]

    # Add some perturbations within each box
    pert = np.random.uniform(-1.0, 1.0, (n_pts, dim)) / float(2 * n_pts)
    X += pert
    return X


def compute_hamming_dist(x1, x2, categorical_dims, normalize=False):
    """
    Compute the hamming distance of two one-hot encoded strings.
    :param x1:
    :param x2:
    :param categorical_dims: list of lists. e.g.
    [[1, 2], [3, 4, 5, 6]] where idx 1 and 2 correspond to the first variable, and
    3, 4, 5, 6 coresponds to the second variable with 4 possible options
    :return:
    """
    dist = 0.
    for i, var_groups in enumerate(categorical_dims):
        if not np.all(x1[var_groups] == x2[var_groups]):
            dist += 1.
    if normalize:
        dist /= len(categorical_dims)
    return dist


def compute_hamming_dist_ordinal(x1, x2, n_categories=None, normalize=False):
    """Same as above, but on ordinal representations."""
    hamming = (x1 != x2).sum()
    if normalize:
        return hamming / len(x1)
    return hamming


def sample_neighbour(x, categorical_dims):
    """Sample a neighbour (i.e. of unnormalised Hamming distance of 1) from x"""
    x_pert = deepcopy(x)
    # Sample a variable where x_pert will differ from the selected sample
    # random.seed(random.randint(0, 1e6))
    choice = random.randint(0, len(categorical_dims) - 1)
    # Change the value of that variable randomly
    var_group = categorical_dims[choice]
    # Confirm what is value of the selected variable in x (we will not sample this point again)
    for var in var_group:
        if x_pert[var] != 0:
            break
    value_choice = random.choice(var_group)
    while value_choice == var:
        value_choice = random.choice(var_group)
    x_pert[var] = 0
    x_pert[value_choice] = 1
    return x_pert


def sample_neighbour_ordinal(x, n_categories):
    """Same as above, but the variables are represented ordinally."""
    x_pert = deepcopy(x)
    # Chooose a variable to modify
    choice = random.randint(0, len(n_categories) - 1)
    # Obtain the current value.
    curr_val = x[choice]
    options = [i for i in range(n_categories[choice]) if i != curr_val]
    x_pert[choice] = random.choice(options)
    return x_pert


def random_sample_within_discrete_tr(x_center, max_hamming_dist, categorical_dims,
                                     mode='ordinal'):
    """Randomly sample a point within the discrete trust region"""
    if max_hamming_dist < 1:  # Normalised hamming distance is used
        bit_change = int(max_hamming_dist * len(categorical_dims))
    else:  # Hamming distance is not normalized
        max_hamming_dist = min(max_hamming_dist, len(categorical_dims))
        bit_change = int(max_hamming_dist)

    x_pert = deepcopy(x_center)
    # Randomly sample n bits to change.
    modified_bits = random.sample(range(len(categorical_dims)), bit_change)
    for bit in modified_bits:
        n_values = len(categorical_dims[bit])
        # Change this value
        selected_value = random.choice(range(n_values))
        # Change to one-hot encoding
        substitute_values = np.array([1 if i == selected_value else 0 for i in range(n_values)])
        x_pert[categorical_dims[bit]] = substitute_values
    return x_pert


def random_sample_within_discrete_tr_ordinal(x_center, max_hamming_dist, n_categories):
    """Same as above, but here we assume an ordinal representation of the categorical variables."""
    # random.seed(random.randint(0, 1e6))
    if max_hamming_dist < 1:
        bit_change = int(max(max_hamming_dist * len(n_categories), 1))
    else:
        bit_change = int(min(max_hamming_dist, len(n_categories)))
    x_pert = deepcopy(x_center)
    modified_bits = random.sample(range(len(n_categories)), bit_change)
    for bit in modified_bits:
        options = np.arange(n_categories[bit])
        x_pert[bit] = int(random.choice(options))
    return x_pert


def local_search(x_center, f: Callable,
                 config,
                 max_hamming_dist,
                 n_restart: int = 1,
                 batch_size: int = 1,
                 step: int = 200):
    """
    Local search algorithm
    :param n_restart: number of restarts
    :param config:
    :param x0: the initial point to start the search
    :param x_center: the center of the trust region. In this case, this should be the optimum encountered so far.
    :param f: the function handle to evaluate x on (the acquisition function, in this case)
    :param max_hamming_dist: maximum Hamming distance from x_center
    :param step: number of maximum local search steps the algorithm is allowed to take.
    :return:
    """

    def _ls(hamming):
        """One restart of local search"""
        # x0 = deepcopy(x_center)
        x_center_local = deepcopy(x_center)
        tol = 100
        trajectory = np.array([x_center_local])
        x = x_center_local

        acq_x = f(x).detach().numpy()
        for i in range(step):
            tol_ = tol
            is_valid = False
            while not is_valid:
                neighbour = sample_neighbour_ordinal(x, config)
                if 0 < compute_hamming_dist_ordinal(x_center_local, neighbour, config) <= hamming \
                        and not any(np.equal(trajectory, neighbour).all(1)):
                    is_valid = True
                else:
                    tol_ -= 1
            if tol_ < 0:
                logging.info("Tolerance exhausted on this local search thread.")
                return x, acq_x

            acq_x = f(x).detach().numpy()
            acq_neighbour = f(neighbour).detach().numpy()
            # print(acq_x, acq_neighbour)

            if acq_neighbour > acq_x:
                logging.info(''.join([str(int(i)) for i in neighbour.flatten()]) + ' ' + str(acq_neighbour))
                x = deepcopy(neighbour)
                # trajectory = np.vstack((trajectory, deepcopy(x)))
        logging.info('local search thread ended with highest acquisition %s' % acq_x)
        # print(compute_hamming_dist_ordinal(x_center, x, n_categories), acq_x)
        # print(x_center)
        return x, acq_x

    X = []
    fX = []
    for i in range(n_restart):
        res = _ls(max_hamming_dist)
        X.append(res[0])
        fX.append(res[1])

    top_idices = np.argpartition(np.array(fX).flatten(), -batch_size)[-batch_size:]
    # select the top-k smallest
    # top_idices = np.argpartition(np.array(fX).flatten(), batch_size)[:batch_size]
    # print(np.array(fX).flatten()[top_idices])
    return np.array([x for i, x in enumerate(X) if i in top_idices]), np.array(fX).flatten()[top_idices]


def interleaved_search(x_center, f: Callable,
                       cat_dims,
                       cont_dims,
                       config,
                       ub,
                       lb,
                       max_hamming_dist,
                       n_restart: int = 1,
                       batch_size: int = 1,
                       interval: int = 1,
                       step: int = 200):
    """
    Interleaved search combining both first-order gradient-based method on the continuous variables and the local search
    for the categorical variables.
    Parameters
    ----------
    x_center: the starting point of the search
    cat_dims: the indices of the categorical dimensions
    cont_dims: the indices of the continuous dimensions
    f: function handle (normally this should be the acquisition function)
    config: the config for the categorical variables
    lb: lower bounds (trust region boundary) for the continuous variables
    ub: upper bounds (trust region boundary) for the continuous variables
    max_hamming_dist: maximum hamming distance boundary (for the categorical variables)
    n_restart: number of restarts of the optimisaiton
    batch_size:
    interval: number of steps to switch over (to start with, we optimise with n_interval steps on the continuous
        variables via a first-order optimiser, then we switch to categorical variables (with the continuous ones fixed)
        and etc.
    step: maximum number of search allowed.

    Returns
    -------

    """
    # todo: the batch setting needs to be changed. For the continuous dimensions, we cannot simply do top-n indices.

    from torch.quasirandom import SobolEngine

    # select the initialising points for both the continuous and categorical variables and then hstack them together
    # x0_cat = np.array([deepcopy(sample_neighbour_ordinal(x_center[cat_dims], config)) for _ in range(n_restart)])
    x0_cat = np.array([deepcopy(random_sample_within_discrete_tr_ordinal(x_center[cat_dims], max_hamming_dist, config))
                       for _ in range(n_restart)])
    # x0_cat = np.array([deepcopy(x_center[cat_dims]) for _ in range(n_restart)])
    seed = np.random.randint(int(1e6))
    sobol = SobolEngine(len(cont_dims), scramble=True, seed=seed)
    x0_cont = sobol.draw(n_restart).cpu().detach().numpy()
    x0_cont = lb + (ub - lb) * x0_cont
    x0 = np.hstack((x0_cat, x0_cont))
    tol = 100
    lb, ub = torch.tensor(lb, dtype=torch.float32), torch.tensor(ub, dtype=torch.float32)

    def _interleaved_search(x0):
        x = deepcopy(x0)
        acq_x = f(x).detach().numpy()
        x_cat, x_cont = x[cat_dims], x[cont_dims]
        n_step = 0
        while n_step <= step:
            # First optimise the continuous part, freezing the categorical part
            def f_cont(x_cont_):
                """The function handle for continuous optimisation"""
                x_ = torch.cat((x_cat_torch, x_cont_)).float()
                return -f(x_)

            x_cont_torch = torch.tensor(x_cont, dtype=torch.float32).requires_grad_(True)
            x_cat_torch = torch.tensor(x_cat, dtype=torch.float32)
            optimizer = torch.optim.Adam([{"params": x_cont_torch}], lr=0.1)
            for _ in range(interval):
                optimizer.zero_grad()
                acq = f_cont(x_cont_torch).float()
                try:
                    acq.backward()
                    # print(x_cont_torch, acq, x_cont_torch.grad)
                    optimizer.step()
                except RuntimeError:
                    print('Exception occured during backpropagation. NaN encountered?')
                    pass
                with torch.no_grad():
                    # Ugly way to do clipping
                    x_cont_torch.data = torch.max(torch.min(x_cont_torch, ub), lb)

            x_cont = x_cont_torch.detach().numpy()
            del x_cont_torch

            # Then freeze the continuous part and optimise the categorical part
            for j in range(interval):
                is_valid = False
                tol_ = tol
                while not is_valid:
                    neighbour = sample_neighbour_ordinal(x_cat, config)
                    if 0 <= compute_hamming_dist_ordinal(x_center[cat_dims], neighbour, config) <= max_hamming_dist:
                        is_valid = True
                    else:
                        tol_ -= 1
                if tol_ < 0:
                    logging.info("Tolerance exhausted on this local search thread.")
                    break
                # acq_x = f(np.hstack((x_cat, x_cont))).detach().numpy()
                acq_neighbour = f(np.hstack((neighbour, x_cont))).detach().numpy()
                if acq_neighbour > acq_x:
                    x_cat = deepcopy(neighbour)
                    acq_x = acq_neighbour
            # print(x_cat, x_cont, acq_x)
            n_step += interval

        x = np.hstack((x_cat, x_cont))
        return x, acq_x

    X, fX = [], []
    for i in range(n_restart):
        res = _interleaved_search(x0[i, :])
        X.append(res[0])
        fX.append(res[1])
    top_idices = np.argpartition(np.array(fX).flatten(), -batch_size)[-batch_size:]
    return np.array([x for i, x in enumerate(X) if i in top_idices]), np.array(fX).flatten()[top_idices]



from copy import deepcopy

import numpy as np
import scipy.stats as ss
from SMKBO.cas.localbo_cat import CASMOPOLITANCat
from SMKBO.cas.localbo_utils import from_unit_cube, latin_hypercube, onehot2ordinal,\
    random_sample_within_discrete_tr_ordinal
import torch


def order_stats(X):
    _, idx, cnt = np.unique(X, return_inverse=True, return_counts=True)
    obs = np.cumsum(cnt)  # Need to do it this way due to ties
    o_stats = obs[idx]
    return o_stats


def copula_standardize(X):
    X = np.nan_to_num(np.asarray(X))  # Replace inf by something large
    assert X.ndim == 1 and np.all(np.isfinite(X))
    o_stats = order_stats(X)
    quantile = np.true_divide(o_stats, len(X) + 1)
    X_ss = ss.norm.ppf(quantile)
    return X_ss


class Optimizer:

    def __init__(self, config,
                 n_init: int = None,
                 wrap_discrete: bool = True,
                 guided_restart: bool = True,
                 **kwargs):
        """Build wrapper class to use an optimizer in benchmark.

        Parameters
        ----------
        config: list. e.g. [2, 3, 4, 5] -- denotes there are 4 categorical variables, with numbers of categories
            being 2, 3, 4, and 5 respectively.
        guided_restart: whether to fit an auxiliary GP over the best points encountered in all previous restarts, and
            sample the points with maximum variance for the next restart.
        global_bo: whether to use the global version of the discrete GP without local modelling
        """

        # Maps the input order.
        self.config = config.astype(int)
        self.true_dim = len(config)
        self.kwargs = kwargs
        # Number of one hot dimensions
        self.n_onehot = int(np.sum(config))
        # One-hot bounds
        self.lb = np.zeros(self.n_onehot)
        self.ub = np.ones(self.n_onehot)
        self.dim = len(self.lb)
        # True dim is simply th`e number of parameters (do not care about one-hot encoding etc).
        self.max_evals = np.iinfo(np.int32).max  # NOTE: Largest possible int
        self.batch_size = None
        self.history = []
        self.wrap_discrete = wrap_discrete
        self.cat_dims = self.get_dim_info(config)

        self.smkbo = CASMOPOLITANCat(
            dim=self.true_dim,
            n_init=n_init if n_init is not None else 2 * self.true_dim + 1,
            max_evals=self.max_evals,
            batch_size=1,  # We need to update this later
            verbose=False,
            config=self.config,
            **kwargs
        )

        # Our modification: define an auxiliary GP
        self.guided_restart = guided_restart
        # keep track of the best X and fX in each restart
        self.best_X_each_restart, self.best_fX_each_restart = None, None
        self.auxiliary_gp = None

    def restart(self):
        from SMKBO.cas.localbo_utils import train_gp

        if self.guided_restart and len(self.smkbo._fX):

            best_idx = self.smkbo._fX.argmin()
            # Obtain the best X and fX within each restart (cas._fX and cas._X get erased at each restart,
            # but cas.X and cas.fX always store the full history
            if self.best_fX_each_restart is None:
                self.best_fX_each_restart = deepcopy(self.smkbo._fX[best_idx])
                self.best_X_each_restart = deepcopy(self.smkbo._X[best_idx])
            else:
                self.best_fX_each_restart = np.vstack((self.best_fX_each_restart, deepcopy(self.smkbo._fX[best_idx])))
                self.best_X_each_restart = np.vstack((self.best_X_each_restart, deepcopy(self.smkbo._X[best_idx])))

            X_tr_torch = torch.tensor(self.best_X_each_restart, dtype=torch.float32).reshape(-1, self.true_dim)
            fX_tr_torch = torch.tensor(self.best_fX_each_restart, dtype=torch.float32).view(-1)
            # Train the auxiliary
            self.auxiliary_gp = train_gp(X_tr_torch, fX_tr_torch, False, 300, )
            # Generate random points in a Thompson-style sampling
            X_init = latin_hypercube(self.smkbo.n_cand, self.dim)
            X_init = from_unit_cube(X_init, self.lb, self.ub)
            if self.wrap_discrete:
                X_init = self.warp_discrete(X_init, )
            X_init = onehot2ordinal(X_init, self.cat_dims)
            with torch.no_grad():
                self.auxiliary_gp.eval()
                X_init_torch = torch.tensor(X_init, dtype=torch.float32)
                # LCB-sampling
                y_cand_mean, y_cand_var = self.auxiliary_gp(
                    X_init_torch).mean.cpu().detach().numpy(), self.auxiliary_gp(
                    X_init_torch).variance.cpu().detach().numpy()
                y_cand = y_cand_mean - 1.96 * np.sqrt(y_cand_var)

            self.X_init = np.ones((self.smkbo.n_init, self.true_dim))
            indbest = np.argmin(y_cand)
            # The initial trust region centre for the new restart
            centre = deepcopy(X_init[indbest, :])
            # The centre is the first point to be evaluated
            self.X_init[0, :] = deepcopy(centre)
            for i in range(1, self.smkbo.n_init):
                # Randomly sample within the initial trust region length around the centre
                self.X_init[i, :] = deepcopy(
                    random_sample_within_discrete_tr_ordinal(centre, self.smkbo.length_init_discrete, self.config))
            self.smkbo._restart()
            self.smkbo._X = np.zeros((0, self.smkbo.dim))
            self.smkbo._fX = np.zeros((0, 1))
            del X_tr_torch, fX_tr_torch, X_init_torch

        else:
            # If guided restart is not enabled, simply sample a number of points equal to the number of evaluated
            self.smkbo._restart()
            self.smkbo._X = np.zeros((0, self.smkbo.dim))
            self.smkbo._fX = np.zeros((0, 1))
            X_init = latin_hypercube(self.smkbo.n_init, self.dim)
            self.X_init = from_unit_cube(X_init, self.lb, self.ub)
            if self.wrap_discrete:
                self.X_init = self.warp_discrete(self.X_init, )
            self.X_init = onehot2ordinal(self.X_init, self.cat_dims)

    def suggest(self, n_suggestions=1):
        if self.batch_size is None:  # Remember the batch size on the first call to suggest
            self.batch_size = n_suggestions
            self.smkbo.batch_size = n_suggestions
            # self.cas.failtol = np.ceil(np.max([4.0 / self.batch_size, self.dim / self.batch_size]))
            self.smkbo.n_init = max([self.smkbo.n_init, self.batch_size])
            self.restart()

        X_next = np.zeros((n_suggestions, self.true_dim))

        # Pick from the initial points
        n_init = min(len(self.X_init), n_suggestions)
        if n_init > 0:
            X_next[:n_init] = deepcopy(self.X_init[:n_init, :])
            self.X_init = self.X_init[n_init:, :]  # Remove these pending points

        # Get remaining points from TuRBO
        n_adapt = n_suggestions - n_init
        if n_adapt > 0:
            if len(self.smkbo._X) > 0:  # Use random points if we can't fit a GP
                X = deepcopy(self.smkbo._X)
                fX = copula_standardize(deepcopy(self.smkbo._fX).ravel())  # Use Copula
                X_next[-n_adapt:, :] = self.smkbo._create_and_select_candidates(X, fX,
                                                                                       length=self.smkbo.length_discrete,
                                                                                       n_training_steps=300,
                                                                                       hypers={})[-n_adapt:, :]

        suggestions = X_next
        return suggestions

    def observe(self, X, y):
        """Send an observation of a suggestion back to the optimizer.

        Parameters
        ----------
        X : list of dict-like
            Places where the objective function has already been evaluated.
            Each suggestion is a dictionary where each key corresponds to a
            parameter being optimized.
        y : array-like, shape (n,)
            Corresponding values where objective has been evaluated
        """
        assert len(X) == len(y)
        # XX = torch.cat([ordinal2onehot(x, self.n_categories) for x in X]).reshape(len(X), -1)
        XX = X
        yy = np.array(y)[:, None]
        # if self.wrap_discrete:
        #     XX = self.warp_discrete(XX, )

        if len(self.smkbo._fX) >= self.smkbo.n_init:
            self.smkbo._adjust_length(yy)

        self.smkbo.n_evals += self.batch_size
        self.smkbo._X = np.vstack((self.smkbo._X, deepcopy(XX)))
        self.smkbo._fX = np.vstack((self.smkbo._fX, deepcopy(yy.reshape(-1, 1))))
        self.smkbo.X = np.vstack((self.smkbo.X, deepcopy(XX)))
        self.smkbo.fX = np.vstack((self.smkbo.fX, deepcopy(yy.reshape(-1, 1))))

        # Check for a restart
        if self.smkbo.length <= self.smkbo.length_min or self.smkbo.length_discrete <= self.smkbo.length_min_discrete:
            self.restart()

    def warp_discrete(self, X, ):

        X_ = np.copy(X)
        # Process the integer dimensions
        if self.cat_dims is not None:
            for categorical_groups in self.cat_dims:
                max_col = np.argmax(X[:, categorical_groups], axis=1)
                X_[:, categorical_groups] = 0
                for idx, g in enumerate(max_col):
                    X_[idx, categorical_groups[g]] = 1
        return X_

    def get_dim_info(self, n_categories):
        dim_info = []
        offset = 0
        for i, cat in enumerate(n_categories):
            dim_info.append(list(range(offset, offset + cat)))
            offset += cat
        return dim_info



import torch
import math
from gpytorch.kernels import Kernel
from gpytorch.constraints import Positive, Interval
from gpytorch.priors import Prior
from typing import Optional, Union


class SincKernel(Kernel):
    r"""
    Implements the Sinc kernel from "Band-Limited Gaussian Processes: The Sinc Kernel" (Tobar 2019).

    The kernel is defined as:
    K(t) = variance * sinc(bandwidth * t) * cos(2 * pi * center_freq * t)

    where:
    - sinc(x) = sin(πx)/(πx)
    - bandwidth controls the frequency band width (Δ in paper)
    - center_freq controls the central frequency (ξ₀ in paper)
    - variance controls the overall scale (σ² in paper)
    """

    def __init__(
            self,
            bandwidth: float = 1.0,
            center_freq: float = 0.5,
            variance: float = 1.0,
            bandwidth_prior: Optional[Prior] = None,
            bandwidth_constraint: Optional[Interval] = None,
            center_freq_prior: Optional[Prior] = None,
            center_freq_constraint: Optional[Interval] = None,
            variance_prior: Optional[Prior] = None,
            variance_constraint: Optional[Interval] = None,
            **kwargs
    ):
        super().__init__(**kwargs)

        # Set default parameter constraints if not provided
        if bandwidth_constraint is None:
            bandwidth_constraint = Interval(1e-3, 1e3)  # Prevents numerical instability
        if center_freq_constraint is None:
            center_freq_constraint = Positive()  # Ensures center frequency is non-negative
        if variance_constraint is None:
            variance_constraint = Positive()  # Ensures variance is positive

        # Register learnable parameters
        self.register_parameter("raw_bandwidth", torch.nn.Parameter(torch.tensor(float(bandwidth))))
        self.register_parameter("raw_center_freq", torch.nn.Parameter(torch.tensor(float(center_freq))))
        self.register_parameter("raw_variance", torch.nn.Parameter(torch.tensor(float(variance))))

        # Apply constraints
        self.register_constraint("raw_bandwidth", bandwidth_constraint)
        self.register_constraint("raw_center_freq", center_freq_constraint)
        self.register_constraint("raw_variance", variance_constraint)

        # Register priors (if provided)
        if bandwidth_prior is not None:
            self.register_prior("bandwidth_prior", bandwidth_prior, lambda: self.bandwidth,
                                lambda v: self._set_bandwidth(v))
        if center_freq_prior is not None:
            self.register_prior("center_freq_prior", center_freq_prior, lambda: self.center_freq,
                                lambda v: self._set_center_freq(v))
        if variance_prior is not None:
            self.register_prior("variance_prior", variance_prior, lambda: self.variance,
                                lambda v: self._set_variance(v))

    # Property accessors for kernel parameters
    @property
    def bandwidth(self) -> torch.Tensor:
        return self.raw_bandwidth_constraint.transform(self.raw_bandwidth)

    def _set_bandwidth(self, value: Union[float, torch.Tensor]) -> None:
        if not torch.is_tensor(value):
            value = torch.tensor(value)
        self.initialize(raw_bandwidth=self.raw_bandwidth_constraint.inverse_transform(value))

    @property
    def center_freq(self) -> torch.Tensor:
        return self.raw_center_freq_constraint.transform(self.raw_center_freq)

    def _set_center_freq(self, value: Union[float, torch.Tensor]) -> None:
        if not torch.is_tensor(value):
            value = torch.tensor(value)
        self.initialize(raw_center_freq=self.raw_center_freq_constraint.inverse_transform(value))

    @property
    def variance(self) -> torch.Tensor:
        return self.raw_variance_constraint.transform(self.raw_variance)

    def _set_variance(self, value: Union[float, torch.Tensor]) -> None:
        if not torch.is_tensor(value):
            value = torch.tensor(value)
        self.initialize(raw_variance=self.raw_variance_constraint.inverse_transform(value))

    def forward(
            self,
            x1: torch.Tensor,
            x2: torch.Tensor,
            diag: bool = False,
            last_dim_is_batch: bool = False,
            **params
    ) -> torch.Tensor:
        """
        Compute the kernel matrix between x1 and x2.

        Args:
            x1 (torch.Tensor): Input tensor of shape (batch_size, N, d)
            x2 (torch.Tensor): Input tensor of shape (batch_size, M, d)
            diag (bool): If True, return only the diagonal of the kernel matrix.
            last_dim_is_batch (bool): Unused, but required for compatibility.

        Returns:
            torch.Tensor: Kernel matrix of shape (batch_size, N, M)
        """

        # Ensure correct input shape (if 1D, add extra dimension)
        if x1.ndim == 1:
            x1 = x1.unsqueeze(-1)
        if x2.ndim == 1:
            x2 = x2.unsqueeze(-1)

        assert x1.size(-1) == x2.size(-1), "Feature dimension mismatch"

        # Compute pairwise L1 distance to ensure shape consistency
        t = torch.cdist(x1, x2, p=1)  # Shape: (batch_size, N, M)

        # Ensure numerical stability by clamping bandwidth away from zero
        bandwidth_t = self.bandwidth.clamp(min=1e-6) * t

        # Compute sinc function with numerical stability for small values
        sinc_val = torch.where(
            bandwidth_t < 1e-4,
            1 - (math.pi ** 2 * bandwidth_t ** 2) / 6,  # Taylor series approximation for sinc(x) near 0
            torch.sinc(bandwidth_t / math.pi)
        )

        # Compute cosine modulation term
        cos_val = torch.cos(2 * math.pi * self.center_freq * t)

        # Compute final kernel matrix
        K = self.variance * sinc_val * cos_val

        # Add small jitter for numerical stability when x1 == x2
        if not diag and torch.equal(x1, x2):
            K = K + torch.eye(K.size(-1), device=K.device) * 1e-6

        return K.diagonal(dim1=-2, dim2=-1) if diag else K

    def __repr__(self) -> str:
        return (
            f"SincKernel(bandwidth={self.bandwidth:.3g}, "
            f"center_freq={self.center_freq:.3g}, "
            f"variance={self.variance:.3g})"
        )



# Implementation of various kernels

from gpytorch.kernels import Kernel, AdditiveKernel
from gpytorch.kernels.matern_kernel import MaternKernel
from gpytorch.kernels.rbf_kernel import RBFKernel
from gpytorch.constraints import Interval
from gpytorch.kernels.spectral_mixture_kernel import SpectralMixtureKernel
from SMKBO.spectral import CauchyMixtureKernel
import torch


class MixtureKernel(Kernel):
    """
    The implementation of the mixed categorical and continuous kernel first proposed in CoCaBO, but re-implemented
    in gpytorch.

    Note that gpytorch uses the pytorch autodiff engine, and there is no need to manually define the derivatives of
    the kernel hyperparameters w.r.t the log-marinal likelihood as in the gpy implementation.
    """
    has_lengthscale = True

    def __init__(self, categorical_dims,
                 continuous_dims,
                 integer_dims=None,
                 lamda=0.5,
                 categorical_kern_type='transformed_overlap',
                 continuous_kern_type='smk',
                 categorical_lengthscale_constraint=None,
                 continuous_lengthscale_constraint=None,
                 categorical_ard=True,
                 continuous_ard=True,
                 num_mixtures1=5,
                 num_mixtures2=4,
                 **kwargs):
        """

        Parameters
        ----------
        categorical_dims: the dimension indices that are categorical/discrete
        continuous_dims: the dimension indices that are continuous
        integer_dims: the **continuous indices** that additionally require integer constraint.
        lamda: \in [0, 1]. The trade-off between product and additive kernels. If this argument is not supplied, then
            lambda will be optimised as if it is an additional kernel hyperparameter
        categorical_kern_type: 'overlap', 'type2'
        continuous_kern_type: 'rbf' or 'mat52' (Matern 5/2)
        categorical_lengthscale_constraint: if supplied, the constraint on the lengthscale of the categorical kernel
        continuous_lengthscale_constraint: if supplied the constraint on the lengthscale of the continuous kernel
        categorical_ard: bool: whether to use Automatic Relevance Determination (ARD) for categorical dimensions
        continuous_ard: bool: whether to use ARD for continouous dimensions
        kwargs: additional parameters.
        """
        super(MixtureKernel, self).__init__(has_lengthscale=True, **kwargs)
        self.optimize_lamda = lamda is None
        self.fixed_lamda = lamda if not self.optimize_lamda else None
        self.categorical_dims = categorical_dims
        self.continuous_dims = continuous_dims
        if integer_dims is not None:
            integer_dims_np = np.asarray(integer_dims).flatten()
            cont_dims_np = np.asarray(self.continuous_dims).flatten()
            if not np.all(np.in1d(integer_dims_np, cont_dims_np)):
                raise ValueError("if supplied, all continuous dimensions with integer constraint must be themselves "
                                 "contained in the continuous_dimensions!")
            # Convert the integer dims in terms of indices of the continous dims
            integer_dims = np.where(np.in1d(self.continuous_dims, integer_dims))[0]

        self.register_parameter(name='raw_lamda', parameter=torch.nn.Parameter(torch.ones(1)))
        # The lambda must be between 0 and 1.
        self.register_constraint('raw_lamda', Interval(0., 1.))

        # Initialise the
        if categorical_kern_type == 'overlap':
            self.categorical_kern = CategoricalOverlap(lengthscale_constraint=categorical_lengthscale_constraint,
                                                       ard_num_dims=len(categorical_dims) if categorical_ard else None)
        elif categorical_kern_type == 'transformed_overlap':
            self.categorical_kern = TransformedCategorical(lengthscale_constraint=categorical_lengthscale_constraint,
                                                           ard_num_dims=len(
                                                                categorical_dims) if categorical_ard else None)
        else:
            raise NotImplementedError("categorical kernel type %s is not implemented. " % categorical_kern_type)

        # By default, we use the Matern 5/2 kernel
        if continuous_kern_type == 'mat52':
            self.continuous_kern = WrappedMatern(nu=2.5, ard_num_dims=len(continuous_dims) if continuous_ard else None,
                                                 integer_dims=integer_dims,
                                                 lengthscale_constraint=continuous_lengthscale_constraint)
        elif continuous_kern_type == 'rbf':
            self.continuous_kern = WrappedRBF(ard_num_dims=len(continuous_dims) if continuous_ard else None,
                                              integer_dims=integer_dims,
                                              lengthscale_constraint=continuous_lengthscale_constraint)
        elif continuous_kern_type == 'smk':
            self.continuous_kern = WrappedSMK(ard_num_dims=len(continuous_dims),
                                              integer_dims=integer_dims, num_mixtures1=num_mixtures1, num_mixtures2=num_mixtures2)
        else:
            raise NotImplementedError("continuous kernel type %s is not implemented. " % continuous_kern_type)

    @property
    def lamda(self):
        if self.optimize_lamda:
            return self.raw_lamda_constraint.transform(self.raw_lamda)
        else:
            return self.fixed_lamda

    @lamda.setter
    def lamda(self, value):
        self._set_lamda(value)

    def _set_lamda(self, value):
        if self.optimize_lamda:
            if not isinstance(value, torch.Tensor):
                value = torch.as_tensor(value).to(self.raw_lamda)
            self.initialize(raw_lamda=self.raw_lamda_constraint.inverse_transform(value))
        else:
            # Manually restrict the value of lamda between 0 and 1.
            if value <= 0:
                self.fixed_lamda = 0.
            elif value >= 1:
                self.fixed_lamda = 1.
            else:
                self.fixed_lamda = value

    def forward(self, x1, x2, diag=False,
                x1_cont=None, x2_cont=None, **params):
        """
        Note that here I also give options to pass the categorical and continuous inputs separately (instead of jointly)
        because the categorical dimensions will not be differentiable, and thus there would be problems when we optimize
        the acquisition function.

        When passed separately, x1 and x2 refer the categorical (non-differentiable) data, whereas x1_cont and x2_cont
        are the continuous (differentiable) data.
        Parameters
        ----------
        x1
        x2
        diag
        x1_cont
        x2_cont
        params

        Returns
        -------

        """
        if x1_cont is None and x2_cont is None:
            assert x1.shape[1] == len(self.categorical_dims) + len(self.continuous_dims), \
                'dimension mismatch. Expected number of dimensions %d but got %d in x1' % \
                (len(self.categorical_dims) + len(self.continuous_dims), x1.shape[1])
            x1_cont, x2_cont = x1[:, self.continuous_dims], x2[:, self.continuous_dims]
            # the categorical kernels are not differentiable w.r.t inputs, detach them to ensure the computing graph of
            # the autodiff engine is not broken.
            x1_cat, x2_cat = x1[:, self.categorical_dims].detach(), x2[:, self.categorical_dims].detach()
        else:
            assert x1.shape[1] == len(self.categorical_dims)
            assert x1_cont.shape[1] == len(self.continuous_dims)
            x1_cat, x2_cat = x1, x2
        # same in cocabo.
        return (1. - self.lamda) * (self.categorical_kern.forward(x1_cat, x2_cat, diag, **params) +
                                    self.continuous_kern.forward(x1_cont, x2_cont, diag, **params)) + \
               self.lamda * self.categorical_kern.forward(x1_cat, x2_cat, diag, **params) * \
               self.continuous_kern.forward(x1_cont, x2_cont, diag, **params)


def wrap(x1, x2, integer_dims):
    """The wrapping transformation for integer dimensions according to Garrido-Merchán and Hernández-Lobato (2020)."""
    if integer_dims is not None:
        for i in integer_dims:
            x1[:, i] = torch.round(x1[:, i])
            x2[:, i] = torch.round(x2[:, i])
    return x1, x2


class WrappedMatern(MaternKernel):
    """Matern kernels wrapped integer type of inputs according to
    Garrido-Merchán and Hernández-Lobato in
    "Dealing with Categorical and Integer-valued Variables in Bayesian Optimization with Gaussian Processes"

    Note: we deal with the categorical-valued variables using the kernels specifically used to deal with
    categorical variables (instead of the one-hot transformation).
    """

    def __init__(self, integer_dims=None, **kwargs):
        super(WrappedMatern, self).__init__(**kwargs)
        self.integer_dims = integer_dims

    def forward(self, x1, x2, diag=False, **params):
        x1, x2 = wrap(x1, x2, self.integer_dims)
        return super().forward(x1, x2, diag=diag, **params)


class WrappedRBF(RBFKernel, WrappedMatern):
    """Similar to above, but applied to RBF."""

    def __init__(self, integer_dims=None, **kwargs):
        super(WrappedRBF, self).__init__(**kwargs)
        self.integer_dims = integer_dims

    def forward(self, x1, x2, diag=False, **params):
        x1, x2 = wrap(x1, x2, self.integer_dims)
        return super().forward(x1, x2, diag=diag, **params)


class SMKKernel(AdditiveKernel):
    """Composite kernel combining two Spectral Mixture Kernels (SMK1 + SMK2)"""

    def __init__(self, num_mixtures1=5, num_mixtures2=4, **kwargs):
        super().__init__()
        self.smk1 = SpectralMixtureKernel(num_mixtures=num_mixtures1, **kwargs)
        self.smk2 = CauchyMixtureKernel(num_mixtures=num_mixtures2, **kwargs)

    @property
    def kernels(self):
        return [self.smk1, self.smk2]


class WrappedSMK(SMKKernel):
    """Wrapped version of SMKKernel with integer dimension rounding"""

    def __init__(self, integer_dims=None, **kwargs):
        super().__init__(**kwargs)
        self.integer_dims = integer_dims

    def forward(self, x1, x2, diag=False, **params):
        x1, x2 = wrap(x1, x2, self.integer_dims)
        return super().forward(x1, x2, diag=diag, **params)


class CategoricalOverlap(Kernel):
    """Implementation of the categorical overlap kernel.
    This is the most basic form of the categorical kernel that essentially invokes a Kronecker delta function
    between any two elements.
    """

    has_lengthscale = True
    def __init__(self, **kwargs):
        super(CategoricalOverlap, self).__init__(has_lengthscale=True, **kwargs)

    def forward(self, x1, x2, diag=False, last_dim_is_batch=False, **params):
        # First, convert one-hot to ordinal representation

        diff = x1[:, None] - x2[None, :]
        # nonzero location = different cat
        diff[torch.abs(diff) > 1e-5] = 1
        # invert, to now count same cats
        diff1 = torch.logical_not(diff).float()
        if self.ard_num_dims is not None and self.ard_num_dims > 1:
            k_cat = torch.sum(self.lengthscale * diff1, dim=-1) / torch.sum(self.lengthscale)
        else:
            # dividing by number of cat variables to keep this term in range [0,1]
            k_cat = torch.sum(diff1, dim=-1) / x1.shape[1]
        if diag:
            return torch.diag(k_cat).float()
        return k_cat.float()


class TransformedCategorical(CategoricalOverlap):
    """
    Second kind of transformed kernel of form:
    $$ k(x, x') = \exp(\frac{\lambda}{n}) \sum_{i=1}^n [x_i = x'_i] )$$ (if non-ARD)
    or
    $$ k(x, x') = \exp(\frac{1}{n} \sum_{i=1}^n \lambda_i [x_i = x'_i]) $$ if ARD
    """

    has_lengthscale = True

    def forward(self, x1, x2, diag=False, last_dim_is_batch=False, exp='rbf', **params):
        diff = x1[:, None] - x2[None, :]
        diff[torch.abs(diff) > 1e-5] = 1
        diff1 = torch.logical_not(diff).float()

        def rbf(d, ard):
            if ard:
                return torch.exp(torch.sum(d * self.lengthscale, dim=-1) / torch.sum(self.lengthscale))
            else:
                return torch.exp(self.lengthscale * torch.sum(d, dim=-1) / x1.shape[1])

        def mat52(d, ard):
            raise NotImplementedError

        if exp == 'rbf':
            k_cat = rbf(diff1, self.ard_num_dims is not None and self.ard_num_dims > 1)
        elif exp == 'mat52':
            k_cat = mat52(diff1, self.ard_num_dims is not None and self.ard_num_dims > 1)
        else:
            raise ValueError('Exponentiation scheme %s is not recognised!' % exp)
        if diag:
            return torch.diag(k_cat).float()
        return k_cat.float()


class OrdinalKernel(Kernel):
    """
    The ordinal version of TransformedCategorical2 kernel (replace the Kronecker delta with
    the distance metric).
    config: the number of vertices per dimension
    """
    def __init__(self, config, **kwargs):
        super(OrdinalKernel, self).__init__(has_lengthscale=True, **kwargs)
        if not isinstance(config, torch.Tensor):
            config = torch.tensor(config).view(-1)
        self.config = config

    def forward(self, x1, x2, diag=False, last_dim_is_batch=False, **params):
        # expected x1 and x2 are of shape N x D respectively
        diff = (x1[:, None] - x2[None, :]) / self.config
        dist = 1. - torch.abs(diff)
        if self.ard_num_dims is not None and self.ard_num_dims > 1:
            k_cat = torch.exp(
                torch.sum(
                    dist * self.lengthscale, dim=-1
                ) / torch.sum(self.lengthscale)
            )
        else:
            k_cat = torch.exp(
                self.lengthscale * torch.sum(dist, dim=-1) / x1.shape[1]
            )
        if diag:
            return torch.diag(k_cat).float()
        return k_cat.float()


if __name__ == '__main__':
    # Test whether the ordinal kernel is doing ok
    import numpy as np
    import matplotlib.pyplot as plt
    x1 = torch.tensor([[13.,  4.],
        [43., 15.],
        [32., 19.],
        [41.,  9.],
        [47., 44.],
        [48., 21.],
        [15., 24.],
        [20., 13.],
        [36., 46.],
        [19., 17.],
        [35.,  6.],
        [39., 50.],
        [24., 10.],
        [45., 18.],
        [29.,  3.],
        [17., 27.],
        [25., 16.],
        [37., 29.],
        [16.,  2.],
        [ 3., 38.]])

    o = OrdinalKernel(config=[51, 51])
    o.lengthscale = 1.
    K = o.forward(x1, x1).detach().numpy()
    plt.imshow(K)
    plt.colorbar()
    plt.show()



import torch
from .base import TestFunction

class Branin(TestFunction):
    """
    Details: http://www.sfu.ca/~ssurjano/branin.html
    Global optimum: :math:`f(-\\pi,12.275)=0.397887`
    """
    problem_type = 'continuous'
    def __init__(self, normalize=False):
        super(Branin, self).__init__(normalize)
        self.min = 0.397887
        self.minimum = torch.tensor([-torch.pi, 12.275])
        self.dim = 2
        self.lb = -3.0 * torch.ones(2)
        self.ub = 3.0 * torch.ones(2)
        self.int_var = torch.tensor([])
        self.cont_var = torch.arange(0, 2)
        self.info = (
            "2-dimensional Branin function \nGlobal minimum: "
            + "f(-pi, 12.275) = 0.397887"
        )

    def compute(self, x, normalize=False):
        x1 = x[:, 0]
        x2 = x[:, 1]

        t = 1 / (8 * torch.pi)
        s = 10
        r = 6
        c = 5 / torch.pi
        b = 5.1 / (4 * torch.pi ** 2)
        a = 1

        term1 = a * (x2 - b * x1 ** 2 + c * x1 - r) ** 2
        term2 = s * (1 - t) * torch.cos(x1)
        f = term1 + term2 + s

        return f.unsqueeze(1)


class Hartmann3(TestFunction):
    """
    Details: http://www.sfu.ca/~ssurjano/hart3.html
    Global optimum: :math:`f(0.114614,0.555649,0.852547)=-3.86278`

    """
    problem_type = 'continuous'
    def __init__(self, normalize=False):
        super(Hartmann3, self).__init__(normalize)
        self.dim = 3
        self.lb = torch.zeros(3)
        self.ub = torch.ones(3)
        self.int_var = torch.tensor([])
        self.cont_var = torch.arange(0, 3)
        self.min = -3.86278
        self.minimum = torch.tensor([0.114614, 0.555649, 0.852547])
        self.info = "3-dimensional Hartmann function \nGlobal maximum: f(0.114614,0.555649,0.852547) = -3.86278"

    def compute(self, x, normalize=False):
        alpha = torch.tensor([1.0, 1.2, 3.0, 3.2])
        A = torch.tensor([
            [3.0, 10, 30],
            [0.1, 10, 35],
            [3.0, 10, 30],
            [0.1, 10, 35]
        ])
        P = torch.tensor([
            [3689, 1170, 2673],
            [4699, 4387, 7470],
            [1091, 8732, 5547],
            [381, 5743, 8828]
        ]) * 1e-4

        f = torch.zeros((x.shape[0], 1))
        for i in range(4):
            tmp = torch.zeros_like(f)
            for j in range(3):
                tmp += (A[i, j] * (x[:, j] - P[i, j]) ** 2).unsqueeze(1)
            f += alpha[i] * torch.exp(-tmp)
        return -f

class Hartmann6(TestFunction):
    """
    Details: http://www.sfu.ca/~ssurjano/hart6.html
    Global optimum: :math:f(0.201,0.150,0.476,0.275,0.311,0.657)=-3.322
    """
    problem_type = 'continuous'
    def __init__(self, normalize=False):
        super(Hartmann6, self).__init__(normalize)
        self.min = -3.32237
        self.minimum = torch.tensor([0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573])
        self.dim = 6
        self.lb = torch.zeros(6)
        self.ub = torch.ones(6)
        self.int_var = torch.tensor([])
        self.cont_var = torch.arange(0, 6)
        self.info = (
            "6-dimensional Hartmann function \nGlobal optimum: "
            + "f(0.2016,0.15001,0.47687,0.27533,0.31165,0.657) = -3.3223"
        )

    def compute(self, x, normalize=False):
        alpha = torch.tensor([1.0, 1.2, 3.0, 3.2])
        A = torch.tensor(
            [
                [10.0, 3.0, 17.0, 3.5, 1.7, 8.0],
                [0.05, 10.0, 17.0, 0.1, 8.0, 14.0],
                [3.0, 3.5, 1.7, 10.0, 17.0, 8.0],
                [17.0, 8.0, 0.05, 10.0, 0.1, 14.0],
            ]
        )
        P = 1e-4 * torch.tensor(
            [
                [1312.0, 1696.0, 5569.0, 124.0, 8283.0, 5886.0],
                [2329.0, 4135.0, 8307.0, 3736.0, 1004.0, 9991.0],
                [2348.0, 1451.0, 3522.0, 2883.0, 3047.0, 6650.0],
                [4047.0, 8828.0, 8732.0, 5743.0, 1091.0, 381.0],
            ]
        )
        result = torch.zeros(x.shape[0])
        for ii in range(4):
            inner = torch.sum(A[ii, :] * (x - P[ii, :]) ** 2, dim=1)
            result += alpha[ii] * torch.exp(-inner)
        return -result[:, None]





from SMKBO.bo import SpectralBO
from SMKBO.test_func import *
import warnings
warnings.filterwarnings("ignore")

# #%% test1
# f = Func2C(lamda=1e-6, normalize=False)
# max_iters = 30
# batch_size = 1
# bo = SpectralBO(problem_type=f.problem_type, cat_vertices=f.config, cont_lb=f.lb, cont_ub=f.ub,
#            cat_dims=f.categorical_dims, cont_dims=f.continuous_dims,
#                  continuous_kern_type='smk', n_Cauchy=5, n_Gaussian=4, n_init=20, acq_func='ei',
#                  noise_variance=None, ard=True)
#
# for i in range(max_iters):
#     x_next = bo.optim.suggest(batch_size)
#     y_next = f.compute(x_next, normalize=f.normalize)
#     bo.optim.observe(x_next, y_next)
#     if f.normalize:
#         Y = np.array(bo.optim.smkbo.fX) * f.std + f.mean
#     else:
#         Y = np.array(bo.optim.smkbo.fX)
#     if Y[:i].shape[0]:
#         argmin = np.argmin(Y[:i * batch_size])
#         print('Iter %d, Last X %s; \n fX:  %.4f. fX_best: %.4f'
#               % (i, x_next, float(Y[-1]), Y[:i * batch_size][argmin]))
#
# #%% test2
# f = Func3C(lamda=1e-6, normalize=False)
# max_iters = 30
# batch_size = 1
# bo = SpectralBO(problem_type=f.problem_type, cat_vertices=f.config, cont_lb=f.lb, cont_ub=f.ub,
#            cat_dims=f.categorical_dims, cont_dims=f.continuous_dims,
#                  continuous_kern_type='mat52', n_init=20, acq_func='ei',
#                  noise_variance=None, ard=True)
#
# for i in range(max_iters):
#     x_next = bo.optim.suggest(batch_size)
#     y_next = f.compute(x_next, normalize=f.normalize)
#     bo.optim.observe(x_next, y_next)
#     if f.normalize:
#         Y = np.array(bo.optim.smkbo.fX) * f.std + f.mean
#     else:
#         Y = np.array(bo.optim.smkbo.fX)
#     if Y[:i].shape[0]:
#         argmin = np.argmin(Y[:i * batch_size])
#         print('Iter %d, Last X %s; \n fX:  %.4f. fX_best: %.4f'
#               % (i, x_next, float(Y[-1]), Y[:i * batch_size][argmin]))

# #%% test3
# f = QUBO()
# max_iters = 23
# batch_size = 1
# bo = SpectralBO(problem_type=f.problem_type, cat_vertices=f.config,  n_init=20, acq_func='ei',
#                  noise_variance=None, ard=True)
#
# for i in range(max_iters):
#     x_next = bo.optim.suggest()
#     y_next = f.compute(x_next)
#     bo.optim.observe(x_next, y_next)
#     Y = np.array(bo.optim.smkbo.fX)
#     if Y[:i].shape[0]:
#         argmin = np.argmin(Y[:i * batch_size])
#         print('Iter %d, Last X %s; \n fX:  %.4f. fX_best: %.4f'
#               % (i, x_next, float(Y[-1]), Y[:i * batch_size][argmin]))

#%% test4
f = Hartmann3()
max_iters = 23
batch_size = 1
bo = SpectralBO(problem_type=f.problem_type, cont_lb=f.lb, cont_ub=f.ub,
                continuous_kern_type='smk', n_Cauchy=9, n_Gaussian=0, n_init=20, acq_func='ucb',
                 noise_variance=None, ard=True)

for i in range(max_iters):
    x_next = bo.optim.suggest(batch_size)
    y_next = f.compute(x_next)
    bo.optim.observe(x_next, y_next)
    Y = np.array(bo.optim.fX)
    if Y[:i].shape[0]:
        argmin = np.argmin(Y[:i * batch_size])
        print('Iter %d, Last X %s; \n fX:  %.4f. fX_best: %.4f'
              % (i, x_next, float(Y[-1]), Y[:i * batch_size][argmin]))



import numpy as np
from .base import TestFunction


class QUBO(TestFunction):
    """
    Quadratic Unconstrained Binary Optimization problem.
    Minimizes x^T Q x where Q is a random symmetric matrix.
    """
    problem_type = 'categorical'

    def __init__(self, dim=10, normalize=False, seed=None):
        super(QUBO, self).__init__(normalize)
        self.dim = dim
        self.dim_bi = dim
        self.n_vertices = np.full(dim, 2)
        self.config = self.n_vertices
        self.lb = np.zeros(dim)
        self.ub = np.ones(dim)
        np.random.seed(seed)
        self.Q = np.random.randn(dim, dim)
        self.Q = (self.Q + self.Q.T) / 2  # Ensure symmetry

        if self.normalize:
            self.mean, self.std = self.sample_normalize()
        else:
            self.mean, self.std = None, None

    def compute(self, X, normalize=True):
        if X.ndim == 1:
            X = X.reshape(1, -1)
        res = np.array([x.T @ self.Q @ x for x in X])
        if normalize and self.mean is not None:
            res = (res - self.mean) / self.std
        return res

    def sample_normalize(self, size=None):
        if size is None:
            size = 2 * self.dim + 1
        y = []
        for _ in range(size):
            x = np.random.randint(0, 2, size=self.dim)
            y.append(self.compute(x, normalize=False))
        y = np.array(y)
        return np.mean(y), np.std(y)


class TrapFunction(TestFunction):
    """
    Deceptive trap function where all zeros is a local optimum
    and all ones is the global optimum.
    """
    problem_type = 'categorical'

    def __init__(self, dim=20, normalize=False):
        super(TrapFunction, self).__init__(normalize)
        self.dim = dim
        self.dim_bi = dim
        self.n_vertices = np.full(dim, 2)
        self.config = self.n_vertices
        self.lb = np.zeros(dim)
        self.ub = np.ones(dim)

        if self.normalize:
            self.mean, self.std = self.sample_normalize()
        else:
            self.mean, self.std = None, None

    def compute(self, X, normalize=True):
        if X.ndim == 1:
            X = X.reshape(1, -1)
        sum_x = np.sum(X, axis=1)
        res = np.where(
            sum_x == 0, 1.0, np.where(
                sum_x == 1, 0.8, sum_x / self.dim
            )
        )
        if normalize and self.mean is not None:
            res = (res - self.mean) / self.std
        return -res

    def sample_normalize(self, size=None):
        if size is None:
            size = 2 * self.dim + 1
        y = []
        for _ in range(size):
            x = np.random.randint(0, 2, size=self.dim)
            y.append(self.compute(x, normalize=False))
        y = np.array(y)
        return np.mean(y), np.std(y)


class MaxCut(TestFunction):
    """
    Max-Cut problem on a randomly generated graph.
    The weight matrix W is symmetric with random entries.
    """
    problem_type = 'categorical'

    def __init__(self, dim=10, normalize=False, seed=None):
        super(MaxCut, self).__init__(normalize)
        self.dim = dim
        self.dim_bi = dim
        self.n_vertices = np.full(dim, 2)
        self.config = self.n_vertices
        self.lb = np.zeros(dim)
        self.ub = np.ones(dim)
        np.random.seed(seed)
        self.W = np.random.rand(dim, dim)  # Random weight matrix
        self.W = (self.W + self.W.T) / 2   # Symmetrize

        if self.normalize:
            self.mean, self.std = self.sample_normalize()
        else:
            self.mean, self.std = None, None

    def compute(self, X, normalize=True):
        if X.ndim == 1:
            X = X.reshape(1, -1)
        res = np.zeros(X.shape[0])
        for i, x in enumerate(X):
            res[i] = -np.sum(self.W * np.outer(x, (1 - x)))  # Minimize negative cut weight
        if normalize and self.mean is not None:
            res = (res - self.mean) / self.std
        return -res

    def sample_normalize(self, size=None):
        if size is None:
            size = 2 * self.dim + 1
        y = []
        for _ in range(size):
            x = np.random.randint(0, 2, size=self.dim)
            y.append(self.compute(x, normalize=False))
        y = np.array(y)
        return np.mean(y), np.std(y)


from SMKBO.cas.optimizer import Optimizer
from SMKBO.cas.optimizer_mixed import MixedOptimizer
from SMKBO.cas.optimizer_cont import OptimizerCont

class SpectralBO:
    def __init__(self, problem_type='mixed', cat_vertices=None, cont_lb=None, cont_ub=None, cat_dims=None, cont_dims=None,
                 continuous_kern_type='smk', n_Cauchy=5, n_Gaussian=4, n_init=20, acq_func='ei',
                 noise_variance=None, ard=True):
        self.problem_type = problem_type
        self.cat_vertices = cat_vertices
        self.cont_lb = cont_lb
        self.cont_ub = cont_ub
        self.cat_dims = cat_dims
        self.cont_dims = cont_dims
        self.continuous_kern_type = continuous_kern_type
        self.n_Cauchy = n_Cauchy
        self.n_Gaussian = n_Gaussian
        self.n_init = n_init
        self.acq_func = acq_func
        self.ard = ard
        self.noise_variance = noise_variance
        self._sanity_check()

        if problem_type == 'mixed':
            kwargs = {"continuous_kern_type": self.continuous_kern_type, "num_mixtures1": self.n_Cauchy,
                      "num_mixtures2": self.n_Gaussian}
            self.optim = MixedOptimizer(self.cat_vertices, self.cont_lb, self.cont_ub, self.cont_dims, self.cat_dims,
                                   n_init=self.n_init, use_ard=self.ard, acq=self.acq_func, kernel_type='mixed',
                                   noise_variance=self.noise_variance, **kwargs)
        elif problem_type == 'categorical':
            self.optim = Optimizer(self.cat_vertices, n_init=self.n_init, use_ard=self.ard, acq=self.acq_func,
                              kernel_type='transformed_overlap', noise_variance=self.noise_variance)
        else:
            self.optim=OptimizerCont(lb=self.cont_lb, ub=self.cont_ub, continuous_kern_type=self.continuous_kern_type,
                                     n_Cauchy=self.n_Cauchy, n_Gaussian=self.n_Gaussian, acq=self.acq_func, n_init=self.n_init)

    def _sanity_check(self):
        assert self.problem_type in ['categorical', 'continuous', 'mixed'], 'Unknown problem type ' + str(self.problem_type)
        assert self.continuous_kern_type in ['mat52', 'rbf', 'smk'], 'Unknown continuous kernel choice ' + str(self.continuous_kern_type)
        assert self.acq_func in ['ucb', 'ei', 'thompson'], 'Unknown acquisition function choice ' + str(self.acq_func)




r"""
Adapted from Botorch package
"""

from __future__ import annotations

import warnings

import torch
from botorch.models.gpytorch import BatchedMultiOutputGPyTorchModel
from botorch.models.model import FantasizeMixin
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform, Standardize
from botorch.models.utils import validate_input_scaling
from botorch.models.utils.gpytorch_modules import (
    get_gaussian_likelihood_with_lognormal_prior,
)
from botorch.utils.containers import BotorchContainer
from botorch.utils.datasets import SupervisedDataset
from botorch.utils.types import _DefaultType, DEFAULT
from gpytorch.distributions.multivariate_normal import MultivariateNormal
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from gpytorch.likelihoods.likelihood import Likelihood
from gpytorch.means.constant_mean import ConstantMean
from gpytorch.means.mean import Mean
from gpytorch.kernels import MaternKernel, RBFKernel, RQKernel, SpectralDeltaKernel, SpectralMixtureKernel, PeriodicKernel, LinearKernel
from gpytorch.models.exact_gp import ExactGP
from gpytorch.module import Module
from SMKBO.spectral import CauchyMixtureKernel, AdaptiveKernel, SincKernel
from torch import Tensor


class SingleTaskGP(BatchedMultiOutputGPyTorchModel, ExactGP, FantasizeMixin):
    r"""A single-task exact GP model, supporting both known and inferred noise levels.

    A single-task exact GP which, by default, utilizes hyperparameter priors
    from [Hvarfner2024vanilla]_. These priors designed to perform well independently of
    the dimensionality of the problem. Moreover, they suggest a moderately low level of
    noise. Importantly, The model works best when covariates are normalized to the unit
    cube and outcomes are standardized (zero mean, unit variance). For a detailed
    discussion on the hyperparameter priors, see
    https://github.com/pytorch/botorch/discussions/2451.

    This model works in batch mode (each batch having its own hyperparameters).
    When the training observations include multiple outputs, this model will use
    batching to model outputs independently.

    Use this model when you have independent output(s) and all outputs use the
    same training data. If outputs are independent and outputs have different
    training data, use the ModelListGP. When modeling correlations between
    outputs, use the MultiTaskGP.

    An example of a case in which noise levels are known is online
    experimentation, where noise can be measured using the variability of
    different observations from the same arm, or provided by outside software.
    Another use case is simulation optimization, where the evaluation can
    provide variance estimates, perhaps from bootstrapping. In any case, these
    noise levels can be provided to `SingleTaskGP` as `train_Yvar`.

    `SingleTaskGP` can also be used when the observations are known to be
    noise-free. Noise-free observations can be modeled using arbitrarily small
    noise values, such as `train_Yvar=torch.full_like(train_Y, 1e-6)`.

    Example:
        Model with inferred noise levels:

        >>> import torch
        >>> from botorch.models.gp_regression import SingleTaskGP
        >>> from botorch.models.transforms.outcome import Standardize
        >>>
        >>> train_X = torch.rand(20, 2, dtype=torch.float64)
        >>> train_Y = torch.sin(train_X).sum(dim=1, keepdim=True)
        >>> outcome_transform = Standardize(m=1)
        >>> inferred_noise_model = SingleTaskGP(
        ...     train_X, train_Y, outcome_transform=outcome_transform,
        ... )

        Model with a known observation variance of 0.2:

        >>> train_Yvar = torch.full_like(train_Y, 0.2)
        >>> observed_noise_model = SingleTaskGP(
        ...     train_X, train_Y, train_Yvar,
        ...     outcome_transform=outcome_transform,
        ... )

        With noise-free observations:

        >>> train_Yvar = torch.full_like(train_Y, 1e-6)
        >>> noise_free_model = SingleTaskGP(
        ...     train_X, train_Y, train_Yvar,
        ...     outcome_transform=outcome_transform,
        ... )
    """

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Yvar: Tensor | None = None,
        likelihood: Likelihood | None = None,
        covar_module: Module | None = None,
        n_mixture: Tensor | None = None,
        mean_module: Mean | None = None,
        outcome_transform: OutcomeTransform | _DefaultType | None = DEFAULT,
        input_transform: InputTransform | None = None,
        n_mixture1: Tensor | None = None,
        n_mixture2: Tensor | None = None,
    ) -> None:
        r"""
        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            train_Yvar: An optional `batch_shape x n x m` tensor of observed
                measurement noise.
            likelihood: A likelihood. If omitted, use a standard
                `GaussianLikelihood` with inferred noise level if `train_Yvar`
                is None, and a `FixedNoiseGaussianLikelihood` with the given
                noise observations if `train_Yvar` is not None.
            covar_module: The module computing the covariance (Kernel) matrix.
                If omitted, uses an `RBFKernel`.
            mean_module: The mean function to be used. If omitted, use a
                `ConstantMean`.
            outcome_transform: An outcome transform that is applied to the
                training data during instantiation and to the posterior during
                inference (that is, the `Posterior` obtained by calling
                `.posterior` on the model will be on the original scale). We use a
                `Standardize` transform if no `outcome_transform` is specified.
                Pass down `None` to use no outcome transform.
            input_transform: An input transform that is applied in the model's
                forward pass.
        """
        self._validate_tensor_args(X=train_X, Y=train_Y, Yvar=train_Yvar)
        if outcome_transform == DEFAULT:
            outcome_transform = Standardize(
                m=train_Y.shape[-1], batch_shape=train_X.shape[:-2]
            )
        with torch.no_grad():
            transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )
        if outcome_transform is not None:
            train_Y, train_Yvar = outcome_transform(train_Y, train_Yvar)
        # Validate again after applying the transforms
        self._validate_tensor_args(X=transformed_X, Y=train_Y, Yvar=train_Yvar)
        ignore_X_dims = getattr(self, "_ignore_X_dims_scaling_check", None)
        validate_input_scaling(
            train_X=transformed_X,
            train_Y=train_Y,
            train_Yvar=train_Yvar,
            ignore_X_dims=ignore_X_dims,
        )
        self._set_dimensions(train_X=train_X, train_Y=train_Y)
        train_X, train_Y, train_Yvar = self._transform_tensor_args(
            X=train_X, Y=train_Y, Yvar=train_Yvar
        )
        if likelihood is None:
            if train_Yvar is None:
                likelihood = get_gaussian_likelihood_with_lognormal_prior(
                    batch_shape=self._aug_batch_shape
                )
            else:
                likelihood = FixedNoiseGaussianLikelihood(
                    noise=train_Yvar, batch_shape=self._aug_batch_shape
                )
        else:
            self._is_custom_likelihood = True
        ExactGP.__init__(
            self, train_inputs=train_X, train_targets=train_Y, likelihood=likelihood
        )
        if mean_module is None:
            mean_module = ConstantMean(batch_shape=self._aug_batch_shape)
        self.mean_module = mean_module
        if covar_module == "rbf":
            covar = RBFKernel(lengthscale=20)
        elif covar_module == 'mat52':
            covar = MaternKernel(nu=2.5, lengthscale=20)
        elif covar_module == 'rq':
            covar = RQKernel(lengthscale=20)
        elif covar_module == 'pe':
            covar = PeriodicKernel()
        elif covar_module == 'smk':
            if not n_mixture1:
                covar = SpectralMixtureKernel(num_mixtures=n_mixture2, ard_num_dims=transformed_X.shape[-1])
            elif not n_mixture2:
                covar = CauchyMixtureKernel(num_mixtures=n_mixture1, ard_num_dims=transformed_X.shape[-1])
            else:
                covar1 = CauchyMixtureKernel(num_mixtures=n_mixture1, ard_num_dims=transformed_X.shape[-1])
                covar2 = SpectralMixtureKernel(num_mixtures=n_mixture2, ard_num_dims=transformed_X.shape[-1])
                covar = covar1 + covar2
        elif covar_module == 'sdk':
            covar = SpectralDeltaKernel(num_dims=transformed_X.shape[-1])
        elif covar_module == 'sinc':
            covar = SincKernel()
        elif covar_module == 'ada':
            kernel_list = [
                RBFKernel(),
                MaternKernel(nu=2.5),
                LinearKernel(),
                RQKernel(lengthscale=20)
            ]
            covar = AdaptiveKernel(kernel_list)
            covar.set_train_data(train_X, train_Y)
        else:
            print("Unknown kernel type")
        self._subset_batch_dict = {
            "mean_module.raw_constant": -1,
            "covar_module.raw_lengthscale": -3,
        }
        if train_Yvar is None:
            self._subset_batch_dict["likelihood.noise_covar.raw_noise"] = -2
        self.covar_module: Module = covar
        # TODO: Allow subsetting of other covar modules
        if outcome_transform is not None:
            self.outcome_transform = outcome_transform
        if input_transform is not None:
            self.input_transform = input_transform
        self.to(train_X)

    @classmethod
    def construct_inputs(
        cls, training_data: SupervisedDataset, *, task_feature: int | None = None
    ) -> dict[str, BotorchContainer | Tensor]:
        r"""Construct `SingleTaskGP` keyword arguments from a `SupervisedDataset`.

        Args:
            training_data: A `SupervisedDataset`, with attributes `train_X`,
                `train_Y`, and, optionally, `train_Yvar`.
            task_feature: Deprecated and allowed only for backward
                compatibility; ignored.

        Returns:
            A dict of keyword arguments that can be used to initialize a `SingleTaskGP`,
            with keys `train_X`, `train_Y`, and, optionally, `train_Yvar`.
        """
        if task_feature is not None:
            warnings.warn(
                "`task_feature` is deprecated and will be ignored. In the "
                "future, this will be an error.",
                DeprecationWarning,
                stacklevel=2,
            )
        return super().construct_inputs(training_data=training_data)

    def forward(self, x: Tensor) -> MultivariateNormal:
        if self.training:
            x = self.transform_inputs(x)
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)


# Note: There used to be `HeteroskedasticSingleTaskGP` here,
# but due to persistent bugs, it was removed in #2616.



import math
from copy import deepcopy

import gpytorch
import numpy as np
import torch
from torch.quasirandom import SobolEngine

from SMKBO.cas.localbo_cat import CASMOPOLITANCat
from SMKBO.cas.localbo_utils import train_gp
from SMKBO.cas.localbo_utils import random_sample_within_discrete_tr_ordinal


class CASMOPOLITANMixed(CASMOPOLITANCat):
    """

    Parameters
    ----------
    config: the configuration for the categorical dimensions
    cat_dims: the list of indices of dimensions that are categorical
    cont_dims: the list of indices of dimensions that are continuous
    *Note*: in general, you should have the first d_cat dimensions as cat_dims, and the rest as the continuous dims.
    lb : Lower variable bounds of the continuous dimensions, numpy.array, shape (d_cont,).
    ub : Upper variable bounds of the continuous dimensions, numpy.array, shape (d_cont,).
    n_init : Number of initial points (2*dim is recommended), int.
    max_evals : Total evaluation budget, int.
    batch_size : Number of points in each batch, int.
    verbose : If you want to print information about the optimization progress, bool.
    use_ard : If you want to use ARD for the GP kernel.
    max_cholesky_size : Largest number of training points where we use Cholesky, int
    n_training_steps : Number of training steps for learning the GP hypers, int
    min_cuda : We use float64 on the CPU if we have this or fewer datapoints
    device : Device to use for GP fitting ("cpu" or "cuda")
    dtype : Dtype to use for GP fitting ("float32" or "float64")
    """

    def __init__(
            self,
            config,
            cat_dim,
            cont_dim,
            lb,
            ub,
            n_init,
            max_evals,
            batch_size=1,
            int_constrained_dims=None,
            verbose=True,
            use_ard=True,
            max_cholesky_size=2000,
            n_training_steps=50,
            min_cuda=1024,
            device="cpu",
            dtype="float32",
            acq='thompson',
            kernel_type='mixed',
            **kwargs
    ):
        super(CASMOPOLITANMixed, self).__init__(len(cat_dim) + len(cont_dim),
                                                n_init, max_evals, config, batch_size, verbose, use_ard,
                                                max_cholesky_size, n_training_steps, min_cuda,
                                                device, dtype, acq, kernel_type, **kwargs)
        # Very basic input checks
        assert lb.ndim == 1 and ub.ndim == 1
        assert len(lb) == len(ub)
        assert np.all(ub > lb)

        self.kwargs = kwargs
        # Save function information for both the continuous and categorical parts.
        self.cat_dims, self.cont_dims = cat_dim, cont_dim
        self.int_constrained_dims = int_constrained_dims
        # self.n_categories = n_cats
        self.lb = lb
        self.ub = ub

    def _adjust_length(self, fX_next):
        if np.min(fX_next) <= np.min(self._fX) - 1e-3 * math.fabs(np.min(self._fX)):
            self.succcount += 1
            self.failcount = 0
        else:
            self.succcount = 0
            self.failcount += 1

        if self.succcount == self.succtol:  # Expand trust region
            # self.length = min([self.tr_multiplier * self.length, self.length_max])
            # For the Hamming distance-bounded trust region, we additively (instead of multiplicatively) adjust.
            self.length_discrete = int(min(self.length_discrete * self.tr_multiplier, self.length_max_discrete))
            self.length = min(self.length * self.tr_multiplier, self.length_max)
            self.succcount = 0
            print("expand", self.length, self.length_discrete)
        elif self.failcount == self.failtol:  # Shrink trust region
            self.failcount = 0
            # Ditto for shrinking.
            self.length_discrete = int(self.length_discrete / self.tr_multiplier)
            self.length = max(self.length / self.tr_multiplier, self.length_min)
            print("Shrink", self.length, self.length_discrete)

    def _create_and_select_candidates(self, X, fX, length, n_training_steps, hypers, return_acq=False):
        # assert X.min() >= 0.0 and X.max() <= 1.0
        # Figure out what device we are running on
        if len(X) < self.min_cuda:
            device, dtype = torch.device("cpu"), torch.float32
        else:
            device, dtype = self.device, self.dtype
        with gpytorch.settings.max_cholesky_size(self.max_cholesky_size):
            X_torch = torch.tensor(X).to(device=device, dtype=dtype)
            y_torch = torch.tensor(fX).to(device=device, dtype=dtype)
            gp = train_gp(
                train_x=X_torch, train_y=y_torch, use_ard=self.use_ard, num_steps=n_training_steps, hypers=hypers,
                kern=self.kernel_type,
                cat_dims=self.cat_dims, cont_dims=self.cont_dims,
                int_constrained_dims=self.int_constrained_dims,
                noise_variance=self.kwargs['noise_variance'] if 'noise_variance' in self.kwargs else None,
                continuous_kern_type = self.kwargs['continuous_kern_type'],
                num_mixtures1=self.kwargs['num_mixtures1'], num_mixtures2=self.kwargs['num_mixtures2']
            )
            # Save state dict
            hypers = gp.state_dict()

        from .localbo_utils import interleaved_search
        x_center = X[fX.argmin().item(), :][None, :]
        # Compute the trust region boundaries for the continuous variables
        base_kernel = gp.covar_module.base_kernel.continuous_kern

        try:  # RBF/Matern
            weights = base_kernel.lengthscale.cpu().detach().numpy().ravel()

        except:  # SMKKernel
            for sub_kernel in base_kernel.kernels:
                if hasattr(sub_kernel, 'mixture_means'):
                    mixture_means = sub_kernel.mixture_means.mean(dim=0).cpu().detach().numpy()
                    weights = (1.0 / (2 * np.pi * mixture_means)).reshape(-1)  # 确保1D输出
                    break
        weights = weights / weights.mean()  # This will make the next line more stable
        weights = weights / np.prod(np.power(weights, 1.0 / len(weights)))  # We now have weights.prod() = 1
        # Note that the length supplied is the discrete length; here we need to call self.length which is for the
        # continuous variables
        lb = np.clip(x_center[0][self.cont_dims] - weights * self.length / 2.0, 0.0, 1.0)
        ub = np.clip(x_center[0][self.cont_dims] + weights * self.length / 2.0, 0.0, 1.0)

        def thompson(n_cand=5000):
            """Thompson sampling"""
            # Generate n_cand of candidates for the discrete variables, in their trust region
            X_cand_cat = np.array([
                random_sample_within_discrete_tr_ordinal(x_center[0][self.cat_dims], length, self.config)
                for _ in range(n_cand)
            ])

            seed = np.random.randint(int(1e6))
            sobol = SobolEngine(len(self.cont_dims), scramble=True, seed=seed)
            pert = sobol.draw(n_cand).to(dtype=dtype, device=device).cpu().detach().numpy()
            pert = lb + (ub - lb) * pert
            prob_perturb = min(20.0 / len(self.cont_dims), 1.0)
            mask = np.random.rand(n_cand, len(self.cont_dims)) <= prob_perturb
            ind = np.where(np.sum(mask, axis=1) == 0)[0]
            mask[ind, np.random.randint(0, len(self.cont_dims) - 1, size=len(ind))] = 1

            X_cand_cont = x_center[0][self.cont_dims].copy() * np.ones((n_cand, len(self.cont_dims)))
            X_cand_cont[mask] = pert[mask]

            X_cand = np.hstack((X_cand_cat, X_cand_cont))

            # Generate n_cand candidates for the continuous variables, in their trust region
            with torch.no_grad(), gpytorch.settings.max_cholesky_size(self.max_cholesky_size):
                X_cand_torch = torch.tensor(X_cand, dtype=torch.float32)
                y_cand = gp.likelihood(gp(X_cand_torch)).sample(
                    torch.Size([self.batch_size])).t().cpu().detach().numpy()

            # Select the best candidates
            X_next = np.ones((self.batch_size, self.dim))
            y_next = np.ones((self.batch_size, 1))
            for i in range(self.batch_size):
                indbest = np.argmin(y_cand[:, i])
                X_next[i, :] = deepcopy(X_cand[indbest, :])
                y_next[i, :] = deepcopy(y_cand[indbest, i])
                y_cand[indbest, :] = np.inf
            return X_next, y_next

        def _ei(X, augmented=True):
            """Expected improvement (with option to enable augmented EI"""
            from torch.distributions import Normal
            if not isinstance(X, torch.Tensor):
                X = torch.tensor(X, dtype=torch.float32)
            if X.dim() == 1:
                X = X.reshape(1, -1)
            gauss = Normal(torch.zeros(1), torch.ones(1))
            # flip for minimization problems
            with gpytorch.settings.max_cholesky_size(self.max_cholesky_size), \
                    gpytorch.settings.fast_pred_var():
                preds = gp(X)
                mean, std = -preds.mean, preds.stddev
                # use in-fill criterion
                mu_star = -gp.likelihood(gp(torch.tensor(x_center[0].reshape(1, -1), dtype=torch.float32))).mean

            u = (mean - mu_star) / std
            ucdf = gauss.cdf(u)
            updf = torch.exp(gauss.log_prob(u))
            ei = std * updf + (mean - mu_star) * ucdf
            if augmented:
                sigma_n = gp.likelihood.noise
                ei *= (1. - torch.sqrt(torch.tensor(sigma_n)) / torch.sqrt(sigma_n + std ** 2))
            return ei

        def _ucb(X, beta=5.):
            """Upper confidence bound"""
            if not isinstance(X, torch.Tensor):
                X = torch.tensor(X, dtype=torch.float32)
            if X.dim() == 1:
                X = X.reshape(1, -1)
            # Invoked when you supply X in one-hot representations
            # if X.shape[1] == self.dim and self.dim != self.true_dim:
            #     X = onehot2ordinal(X, self.cat_dims)
            with gpytorch.settings.max_cholesky_size(self.max_cholesky_size), \
                    gpytorch.settings.fast_pred_var():
                preds = gp.likelihood(gp(X))
            mean, std = preds.mean, preds.stddev
            return -(mean + beta * std)

        if self.acq in ['ei', 'ucb']:
            if self.batch_size == 1:
                if self.acq == 'ei':
                    X_next, acq_next = interleaved_search(x_center[0], _ei,
                                                          self.cat_dims, self.cont_dims,
                                                          self.config, ub, lb, length, 3, self.batch_size,
                                                          interval=1)
                else:
                    X_next, acq_next = interleaved_search(x_center[0], _ucb,
                                                          self.cat_dims, self.cont_dims,
                                                          self.config, ub, lb, length, 3, self.batch_size,
                                                          interval=1)
            else:
                # batch setting: for these, we use the fantasised points {x, y}
                X_next = torch.tensor([], dtype=torch.float32)
                acq_next = np.array([])
                for p in range(self.batch_size):
                    if self.acq == 'ei':
                        x_next, acq = interleaved_search(x_center[0], _ei,
                                                         self.cat_dims, self.cont_dims,
                                                         self.config, ub, lb, length, 3, 1, interval=1)
                    else:
                        x_next, acq = interleaved_search(x_center[0], _ucb,
                                                         self.cat_dims, self.cont_dims,
                                                         self.config, ub, lb, length, 3, 1, interval=1)
                    x_next = torch.tensor(x_next, dtype=torch.float32)
                    # The fantasy point is filled by the posterior mean of the Gaussian process.
                    y_next = gp(x_next).mean.detach()
                    with gpytorch.settings.max_cholesky_size(self.max_cholesky_size):
                        X_torch = torch.cat((X_torch, x_next), dim=0)
                        y_torch = torch.cat((y_torch, y_next), dim=0)
                        gp = train_gp(
                            train_x=X_torch, train_y=y_torch, use_ard=self.use_ard, num_steps=n_training_steps,
                            hypers=hypers,
                            kern=self.kernel_type,
                            cat_dims=self.cat_dims, cont_dims=self.cont_dims,
                            int_constrained_dims=self.int_constrained_dims,
                            noise_variance=self.kwargs['noise_variance'] if 'noise_variance' in self.kwargs else None,
                            continuous_kern_type=self.kwargs['continuous_kern_type'],
                            num_mixtures1=self.kwargs['num_mixtures1'], num_mixtures2=self.kwargs['num_mixtures2']
                        )
                    X_next = torch.cat((X_next, x_next), dim=0)
                    acq_next = np.hstack((acq_next, acq))

        elif self.acq == 'thompson':
            X_next, acq_next = thompson()
        else:
            raise ValueError('Unknown acquisition function choice %s' % self.acq)

        X_next = np.array(X_next)
        if return_acq:
            return X_next, acq_next
        return X_next



#!/usr/bin/env python3

from typing import Optional, Union, List, Type

import torch
from linear_operator.operators import LinearOperator, MatmulLinearOperator, RootLinearOperator
from torch import Tensor

import gpytorch
from gpytorch.constraints import Interval, Positive
from gpytorch.priors import Prior
from gpytorch.kernels.kernel import Kernel



class AdaptiveKernel(Kernel):
    r"""
    Computes a covariance matrix by adaptively selecting the best kernel from a list of candidates
    based on marginal log likelihood (MLL) during training.

    Args:
        kernel_list: List of kernel classes or instances to choose from.
        ard_num_dims: Set this if you want a separate lengthscale for each input dimension.
        variance_prior: Prior over the variance parameter.
        variance_constraint: Constraint to place on variance parameter.
        active_dims: List of data dimensions to operate on.
        **kwargs: Additional arguments to pass to each kernel initialization.
    """

    def __init__(
            self,
            kernel_list: List[Union[Type[Kernel], Kernel]],
            ard_num_dims: Optional[int] = None,
            **kwargs,
    ):
        super().__init__(**kwargs)
        self.kernels = torch.nn.ModuleList()
        for kernel in kernel_list:
            if isinstance(kernel, Kernel):
                self.kernels.append(kernel)  # Instance
            else:
                self.kernels.append(kernel(  # Initialization from Class
                    ard_num_dims=ard_num_dims,
                    **kwargs
                ))

        self.active_kernel_idx = None
        self._prev_mlls = None

    def select_best_kernel(self, x: Tensor, y: Tensor) -> int:
        """
        Selects the kernel with highest marginal log likelihood given current data.

        Args:
            x: Input training data.
            y: Target training data.

        Returns:
            Index of the best kernel.
        """
        mlls = []
        for kernel in self.kernels:
            # Compute marginal log likelihood for each kernel
            K = kernel(x, x).add_jitter(1e-6)
            try:
                L = K.cholesky()
                mll = gpytorch.distributions.MultivariateNormal(
                    torch.zeros_like(y), L
                ).log_prob(y)
                mlls.append(mll.item())
            except:
                mlls.append(-float('inf'))

        self._prev_mlls = mlls
        return torch.argmax(torch.tensor(mlls)).item()

    @property
    def active_kernel(self) -> Kernel:
        """Returns the currently active kernel."""
        if self.active_kernel_idx is None:
            raise RuntimeError("No active kernel selected. Call forward() first.")
        return self.kernels[self.active_kernel_idx]

    def forward(
            self,
            x1: Tensor,
            x2: Tensor,
            diag: bool = False,
            last_dim_is_batch: bool = False,
            select_kernel: bool = True,
            **params
    ) -> Union[Tensor, LinearOperator]:
        """
        Forward pass that automatically selects best kernel on first call.

        Args:
            select_kernel: If True, will select best kernel before forward pass.
                          Set to False after first call for consistent behavior.
        """
        if select_kernel and (self.active_kernel_idx is None or self.training):
            # During training or first call, select best kernel
            if not hasattr(self, "_train_x") or not hasattr(self, "_train_y"):
                raise RuntimeError("Must set train data before forward pass")
            self.active_kernel_idx = self.select_best_kernel(self._train_x, self._train_y)

        return self.active_kernel(x1, x2, diag=diag, last_dim_is_batch=last_dim_is_batch, **params)

    def set_train_data(self, x: Tensor, y: Tensor):
        """Store training data for kernel selection."""
        self._train_x = x
        self._train_y = y

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"(kernels={[k.__class__.__name__ for k in self.kernels]})"


import numpy as np
from .base import TestFunction


class Func2C(TestFunction):
    """Func2C is a mixed categorical and continuous function. The first 2 dimensions are categorical,
    with possible 3 and 5 possible values respectively. The last 2 dimensions are continuous"""

    """
    Global minimum of this function is at
    x* = [1, 1, -0.0898/2, 0.7126/2]
    with f(x*) = -0.2063
    """
    problem_type = 'mixed'

    def __init__(self, lamda=1e-6, normalize=False):
        # Specifies the indices of the dimensions that are categorical and continuous, respectively
        super(Func2C, self).__init__(normalize)
        self.categorical_dims = np.array([0, 1])
        self.continuous_dims = np.array([2, 3])
        self.dim = len(self.categorical_dims) + len(self.continuous_dims)
        self.n_vertices = np.array([3, 5])
        self.config = self.n_vertices
        # Specfies the range for the continuous variables
        self.lb = np.array([-1, -1])
        self.ub = np.array([1, 1])
        self.lamda = lamda

        if self.normalize:
            self.mean, self.std = self.sample_normalize()
        else:
            self.mean, self.std = None, None

    def compute(self, X, normalize=True):
        if X.ndim == 1:
            X = X.reshape(1, -1)
        N = X.shape[0]
        res = np.zeros((N, ))
        X_cat = X[:, self.categorical_dims]
        X_cont = X[:, self.continuous_dims]
        X_cont = X_cont * 2

        for i, X in enumerate(X):
            if X_cat[i, 0] == 0:
                res[i] = myrosenbrock(X_cont[i, :])
            elif X_cat[i, 0] == 1:
                res[i] = mysixhumpcamp(X_cont[i, :])
            else:
                res[i] = mybeale(X_cont[i, :])

            if X_cat[i, 1] == 0:
                res[i] += myrosenbrock(X_cont[i, :])
            elif X_cat[i, 1] == 1:
                res[i] += mysixhumpcamp(X_cont[i, :])
            else:
                res[i] += mybeale(X_cont[i, :])
        res += self.lamda * np.random.rand(*res.shape)
        return res

    def sample_normalize(self, size=None):
        from SMKBO.cas.localbo_utils import latin_hypercube, from_unit_cube
        if size is None:
            size = 2 * self.dim + 1
        y = []
        for i in range(size):
            x_cat = np.array([np.random.choice(self.config[_]) for _ in range(self.categorical_dims.shape[0])])
            x_cont = latin_hypercube(1, self.continuous_dims.shape[0])
            x_cont = from_unit_cube(x_cont, self.lb, self.ub).flatten()
            x = np.hstack((x_cat, x_cont))
            y.append(self.compute(x, normalize=False))
        y = np.array(y)
        return np.mean(y), np.std(y)


class Func3C(TestFunction):
    """
    Func3C is a simlar function, but with 3 categorical variables (each of which is binary) and 2 continuous variables
    """
    problem_type = 'mixed'

    def __init__(self, lamda=1e-6, normalize=False):
        super(Func3C, self).__init__(normalize)
        self.categorical_dims = np.array([0, 1, 2])
        self.continuous_dims = np.array([3, 4])
        self.dim = len(self.categorical_dims) + len(self.continuous_dims)
        self.n_vertices = np.array([2, 2, 2])
        self.config = self.n_vertices
        # Specfies the range for the continuous variables
        self.lb = np.array([-1, -1])
        self.ub = np.array([1, 1])
        self.lamda = lamda
        if self.normalize:
            self.mean, self.std = self.sample_normalize()
        else:
            self.mean, self.std = None, None

    def compute(self, X, normalize=True):
        if X.ndim == 1:
            X = X.reshape(1, -1)
        N = X.shape[0]
        res = np.zeros((N, ))
        X_cat = X[:, self.categorical_dims]
        X_cont = X[:, self.continuous_dims]
        X_cont = X_cont * 2

        for i, X in enumerate(X):
            if X_cat[i, 0] == 0:
                res[i] = myrosenbrock(X_cont[i, :])
            elif X_cat[i, 0] == 1:
                res[i] = mysixhumpcamp(X_cont[i, :])
            elif X_cat[i, 0] == 2:  # should never be activated
                res[i] = mybeale(X_cont[i, :])

            if X_cat[i, 1] == 0:
                res[i] += myrosenbrock(X_cont[i, :])
            elif X_cat[i, 1] == 1:
                res[i] += mysixhumpcamp(X_cont[i, :])
            else: # should never be activated
                res[i] += mybeale(X_cont[i, :])

            if X_cat[i, 2] == 0:
                res[i] += 5 * mysixhumpcamp(X_cont[i, :])
            elif X_cat[i, 2] == 1:
                res[i] += 2 * myrosenbrock(X_cont[i, :])
            else: # should never be activated
                res[i] += X_cont[i, 2] * mybeale(X_cont[i, :])

        res += self.lamda * np.random.rand(*res.shape)
        return res

    def sample_normalize(self, size=None):
        from SMKBO.cas.localbo_utils import latin_hypercube, from_unit_cube
        if size is None:
            size = 2 * self.dim + 1
        y = []
        for i in range(size):
            x_cat = np.array([np.random.choice(self.config[_]) for _ in range(self.categorical_dims.shape[0])])
            x_cont = latin_hypercube(1, self.continuous_dims.shape[0])
            x_cont = from_unit_cube(x_cont, self.lb, self.ub).flatten()
            x = np.hstack((x_cat, x_cont))
            y.append(self.compute(x, normalize=False))
        y = np.array(y)
        return np.mean(y), np.std(y)


class Ackley53(TestFunction):
    problem_type = 'mixed'

    # Taken and adapted from the the MVRSM codebase
    def __init__(self, lamda=1e-6, normalize=False):
        super(Ackley53, self).__init__(normalize)
        self.categorical_dims = np.arange(0, 50)
        self.continuous_dims = np.array([50, 51, 52])
        self.dim = len(self.continuous_dims) + len(self.categorical_dims)
        self.n_vertices = 2 * np.ones(len(self.categorical_dims), dtype=int)
        self.config = self.n_vertices
        self.lamda = lamda
        # specifies the range for the continuous variables
        self.lb, self.ub = np.array([-1, -1, -1]), np.array([+1, +1, +1])

    @staticmethod
    def _ackley(X):
        a = 20
        b = 0.2
        c = 2 * np.pi
        sum_sq_term = -a * np.exp(-b * np.sqrt(np.sum(np.square(X), axis=1) / 53))
        cos_term = -1 * np.exp(np.sum(np.cos(c * np.copy(X)) / 53, axis=1))
        result = a + np.exp(1) + sum_sq_term + cos_term
        return result

    def compute(self, X, normalize=None):
        if X.ndim == 1:
            X = X.reshape(1, -1)
        # To make sure there is no cheating, round the discrete variables before calling the function
        X[:, self.categorical_dims] = np.round(X[:, self.categorical_dims])
        result = self._ackley(X)
        return result + self.lamda * np.random.rand(*result.shape)


# =============================================================================
# Rosenbrock Function (f_min = 0)
# https://www.sfu.ca/~ssurjano/rosen.html
# =============================================================================
def myrosenbrock(X):
    X = np.asarray(X)
    X = X.reshape((-1, 2))
    if len(X.shape) == 1:  # one observation
        x1 = X[0]
        x2 = X[1]
    else:  # multiple observations
        x1 = X[:, 0]
        x2 = X[:, 1]
    fx = 100 * (x2 - x1 ** 2) ** 2 + (x1 - 1) ** 2
    return fx.reshape(-1, 1) / 300


# =============================================================================
#  Six-hump Camel Function (f_min = - 1.0316 )
#  https://www.sfu.ca/~ssurjano/camel6.html
# =============================================================================
def mysixhumpcamp(X):
    X = np.asarray(X)
    X = np.reshape(X, (-1, 2))
    if len(X.shape) == 1:
        x1 = X[0]
        x2 = X[1]
    else:
        x1 = X[:, 0]
        x2 = X[:, 1]
    term1 = (4 - 2.1 * x1 ** 2 + (x1 ** 4) / 3) * x1 ** 2
    term2 = x1 * x2
    term3 = (-4 + 4 * x2 ** 2) * x2 ** 2
    fval = term1 + term2 + term3
    return fval.reshape(-1, 1) / 10


# =============================================================================
# Beale function (f_min = 0)
# https://www.sfu.ca/~ssurjano/beale.html
# =============================================================================
def mybeale(X):
    X = np.asarray(X) / 2
    X = X.reshape((-1, 2))
    if len(X.shape) == 1:
        x1 = X[0] * 2
        x2 = X[1] * 2
    else:
        x1 = X[:, 0] * 2
        x2 = X[:, 1] * 2
    fval = (1.5 - x1 + x1 * x2) ** 2 + (2.25 - x1 + x1 * x2 ** 2) ** 2 + (
            2.625 - x1 + x1 * x2 ** 3) ** 2
    return fval.reshape(-1, 1) / 50


if __name__ == '__main__':
    f = Func3C()
    print(f.sample_normalize(10))



def spearman(pred, target) -> float:
    """Compute the spearman correlation coefficient between prediction and target"""
    from scipy import stats
    coef_val, p_val = stats.spearmanr(pred, target)
    return coef_val


def pearson(pred, target) -> float:
    from scipy import stats
    coef_val, p_val = stats.pearsonr(pred, target)
    return coef_val


def negative_log_likelihood(pred, pred_std, target) -> float:
    """Compute the negative log-likelihood on the validation dataset"""
    from scipy.stats import norm
    import numpy as np
    n = pred.shape[0]
    res = 0.
    for i in range(n):
        res += (
            np.log(norm.pdf(target[i], pred[i], pred_std[i])).sum()
        )
    return -res


def get_dim_info(n_categories):
    dim_info = []
    offset = 0
    for i, cat in enumerate(n_categories):
        dim_info.append(list(range(offset, offset + cat)))
        offset += cat
    return dim_info



# An abstract class implementation for all test functions

from abc import abstractmethod
import numpy as np


class TestFunction:
    """
    The abstract class for all benchmark functions acting as objective functions for BO.
    Note that we assume all problems will be minimization problem, so convert maximisation problems as appropriate.
    """

    # this should be changed if we are tackling a mixed, or continuous problem, for e.g.
    problem_type = 'categorical'

    def __init__(self, normalize=True, **kwargs):
        self.normalize = normalize
        self.n_vertices = None
        self.config = None
        self.dim = None
        self.continuous_dims = None
        self.categorical_dims = None
        self.int_constrained_dims = None

    def _check_int_constrained_dims(self):
        if self.int_constrained_dims is None:
            return
        assert self.continuous_dims is not None, 'int_constrained_dims must be a subset of the continuous_dims, ' \
                                                 'but continuous_dims is not supplied!'
        int_dims_np = np.asarray(self.int_constrained_dims)
        cont_dims_np = np.asarray(self.continuous_dims)
        assert np.all(np.in1d(int_dims_np, cont_dims_np)), "all continuous dimensions with integer " \
                                                           "constraint must be themselves contained in the " \
                                                           "continuous_dimensions!"

    @abstractmethod
    def compute(self, x, normalize=None):
        raise NotImplementedError()

    def sample_normalize(self, size=None):
        if size is None:
            size = 2 * self.dim + 1
        y = []
        for i in range(size):
            x = np.array([np.random.choice(self.config[_]) for _ in range(self.dim)])
            y.append(self.compute(x, normalize=False, ))
        y = np.array(y)
        return np.mean(y), np.std(y)

    def __call__(self, *args, **kwargs):
        return self.compute(*args, **kwargs)


from SMKBO.cas.localbo_mixed import CASMOPOLITANMixed
from SMKBO.cas.optimizer import *
from torch.quasirandom import SobolEngine


class MixedOptimizer(Optimizer):

    def __init__(self, config,
                 lb, ub,
                 cont_dims,
                 cat_dims,
                 int_constrained_dims=None,
                 n_init: int = None,
                 wrap_discrete: bool = True,
                 guided_restart: bool = True,
                 **kwargs):
        """Build wrapper class to use an optimizer in benchmark.

        Parameters
        ----------
        config: list. e.g. [2, 3, 4, 5] -- denotes there are 4 categorical variables, with numbers of categories
            being 2, 3, 4, and 5 respectively.
        guided_restart: whether to fit an auxiliary GP over the best points encountered in all previous restarts, and
            sample the points with maximum variance for the next restart.
        """
        super(MixedOptimizer, self).__init__(config, n_init, wrap_discrete, guided_restart, **kwargs)

        self.kwargs = kwargs
        # Maps the input order.
        self.d_cat, self.d_cont = cat_dims, cont_dims
        self.true_dim = len(cont_dims) + len(cat_dims)

        # Number of one hot dimensions
        self.n_onehot = int(np.sum(config))
        # One-hot bounds
        self.lb = np.hstack((np.zeros(self.n_onehot), lb))
        self.ub = np.hstack((np.ones(self.n_onehot), ub))
        self.dim = len(self.lb)


        self.smkbo = CASMOPOLITANMixed(
            config=self.config,
            cat_dim=cat_dims,
            cont_dim=cont_dims,
            int_constrained_dims=int_constrained_dims,
            lb=lb[self.n_onehot:], ub=ub[self.n_onehot:],
            n_init=n_init if n_init is not None else 2 * self.true_dim + 1,
            max_evals=self.max_evals,
            batch_size=1,  # We need to update this later
            verbose=False,
            **kwargs
        )

    def restart(self):
        from .localbo_utils import train_gp

        if self.guided_restart and len(self.smkbo._fX):
            # Get the best index
            best_idx = self.smkbo._fX.argmin()
            # Obtain the best X and fX within each restart (cas._fX and cas._X get erased at each restart,
            # but cas.X and cas.fX always store the full history
            if self.best_fX_each_restart is None:
                self.best_fX_each_restart = deepcopy(self.smkbo._fX[best_idx])
                self.best_X_each_restart = deepcopy(self.smkbo._X[best_idx])
            else:
                self.best_fX_each_restart = np.vstack((self.best_fX_each_restart, deepcopy(self.smkbo._fX[best_idx])))
                self.best_X_each_restart = np.vstack((self.best_X_each_restart, deepcopy(self.smkbo._X[best_idx])))

            X_tr_torch = torch.tensor(self.best_X_each_restart, dtype=torch.float32).reshape(-1, self.true_dim)
            fX_tr_torch = torch.tensor(self.best_fX_each_restart, dtype=torch.float32).view(-1)
            # Train the auxiliary
            self.auxiliary_gp = train_gp(X_tr_torch, fX_tr_torch, False, 300,
                                         cat_dims=self.d_cat,
                                         cont_dims=self.d_cont,
                                         kern='mixed',
                                         noise_variance=self.kwargs[
                                             'noise_variance'] if 'noise_variance' in self.kwargs else None,
                                         num_mixtures1=self.kwargs['num_mixtures1'],
                                         num_mixtures2=self.kwargs['num_mixtures2']
                                         )
            # Generate random points in a Thompson-style sampling
            X_init = latin_hypercube(self.smkbo.n_cand, self.dim)
            X_init = from_unit_cube(X_init, self.lb, self.ub)
            # Isolate the continuous part and categorical part
            X_init_cat, X_init_cont = X_init[:, :self.n_onehot], X_init[:, self.n_onehot:]
            if self.wrap_discrete:
                X_init_cat = self.warp_discrete(X_init_cat, )
            X_init_cat = onehot2ordinal(X_init_cat, self.cat_dims)
            # Put the two parts back by a hstack
            X_init = np.hstack((X_init_cat, X_init_cont))

            with torch.no_grad():
                self.auxiliary_gp.eval()
                X_init_torch = torch.tensor(X_init, dtype=torch.float32)
                # LCB sampling
                y_cand_mean, y_cand_var = self.auxiliary_gp(
                    X_init_torch).mean.cpu().detach().numpy(), self.auxiliary_gp(
                    X_init_torch).variance.cpu().detach().numpy()
                y_cand = y_cand_mean - 1.96 * np.sqrt(y_cand_var)

                # Maximum variance sampling
                # y_init = self.auxiliary_gp(X_init_torch).variance.cpu().detach().numpy()
            self.X_init = np.ones((self.smkbo.n_init, self.true_dim))
            indbest = np.argmin(y_cand)
            # cThe initial trust region centre for the restart
            centre = deepcopy(X_init[indbest, :])
            # Separate the continuous and categorical parts of the centre.
            centre_cat, centre_cont = centre[self.d_cat], centre[self.d_cont]

            # Generate random samples around the continuous centre similar to the original TuRBO
            centre_cont = centre_cont[None, :]
            lb = np.clip(centre_cont - self.smkbo.length / 2.0, self.lb[self.n_onehot:], self.ub[self.n_onehot:])
            ub = np.clip(centre_cont + self.smkbo.length / 2.0, self.lb[self.n_onehot:], self.ub[self.n_onehot:])
            seed = np.random.randint(int(1e6))
            sobol = SobolEngine(len(self.d_cont), scramble=True, seed=seed)
            pert = sobol.draw(self.smkbo.n_init).to(dtype=torch.float32).cpu().detach().numpy()
            pert = lb + (ub - lb) * pert
            prob_perturb = min(20.0 / len(self.d_cont), 1.0)
            mask = np.random.rand(self.smkbo.n_init, len(self.d_cont)) <= prob_perturb
            ind = np.where(np.sum(mask, axis=1) == 0)[0]
            mask[ind, np.random.randint(0, len(self.d_cont) - 1, size=len(ind))] = 1
            X_init_cont = centre_cont.copy() * np.ones((self.smkbo.n_init, len(self.d_cont)))
            X_init_cont[mask] = pert[mask]

            # Generate the random samples around the categorical centre similar to the discrete CASMOPOLITAN
            X_init_cat = []
            for i in range(self.smkbo.n_init):
                cat_sample = deepcopy(
                    random_sample_within_discrete_tr_ordinal(centre_cat, self.smkbo.length_init_discrete, self.config),
                )
                X_init_cat.append(cat_sample)
            X_init_cat = np.array(X_init_cat)
            self.X_init = np.hstack((X_init_cat, X_init_cont))

            self.smkbo._restart()
            self.smkbo._X = np.zeros((0, self.smkbo.dim))
            self.smkbo._fX = np.zeros((0, 1))
            del X_tr_torch, fX_tr_torch, X_init_torch

        else:
            # If guided restart is not enabled, simply sample a number of points equal to the number of evaluated
            self.smkbo._restart()
            self.smkbo._X = np.zeros((0, self.smkbo.dim))
            self.smkbo._fX = np.zeros((0, 1))
            X_init = latin_hypercube(self.smkbo.n_init, self.dim)
            X_init = from_unit_cube(X_init, self.lb, self.ub)
            X_init_cat, X_init_cont = X_init[:, :self.n_onehot], X_init[:, self.n_onehot:]
            if self.wrap_discrete:
                X_init_cat = self.warp_discrete(X_init_cat, )
            X_init_cat = onehot2ordinal(X_init_cat, self.cat_dims)
            # Put the two parts back by a hstack
            self.X_init = np.hstack((X_init_cat, X_init_cont))

    def suggest(self, n_suggestions=1):
        if self.batch_size is None:  # Remember the batch size on the first call to suggest
            self.batch_size = n_suggestions
            self.smkbo.batch_size = n_suggestions
            self.smkbo.n_init = max([self.smkbo.n_init, self.batch_size])
            self.restart()

        X_next = np.zeros((n_suggestions, self.true_dim))

        # Pick from the initial points
        n_init = min(len(self.X_init), n_suggestions)
        if n_init > 0:
            X_next[:n_init, :] = deepcopy(self.X_init[:n_init, :])
            self.X_init = self.X_init[n_init:, :]  # Remove these pending points

        # Get remaining points from TuRBO
        n_adapt = n_suggestions - n_init
        if n_adapt > 0:
            if len(self.smkbo._X) > 0:  # Use random points if we can't fit a GP
                X = deepcopy(self.smkbo._X)
                # Pre-process the continuous dimensions
                X[:, self.smkbo.cont_dims] = to_unit_cube(X[:, self.smkbo.cont_dims], self.lb[self.n_onehot:],
                                                          self.ub[self.n_onehot:])
                fX = copula_standardize(deepcopy(self.smkbo._fX).ravel())  # Use Copula
                next = self.smkbo._create_and_select_candidates(X, fX, length=self.smkbo.length_discrete,
                                                                       n_training_steps=300, hypers={})[-n_adapt:, :]
                next[:, self.smkbo.cont_dims] = from_unit_cube(next[:, self.smkbo.cont_dims], self.lb[self.n_onehot:],
                                                               self.ub[self.n_onehot:])
                X_next[-n_adapt:, :] = next

                # Unwarp the suggestions
        suggestions = X_next
        return suggestions



import torch

from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.mlls import ExactMarginalLogLikelihood
from gpytorch.constraints.constraints import GreaterThan
from gpytorch.priors.torch_priors import GammaPrior
from botorch.fit import fit_gpytorch_mll
from botorch.acquisition import ExpectedImprovement, ProbabilityOfImprovement, UpperConfidenceBound
from botorch.optim import optimize_acqf

from SMKBO.spectral.gp_regression import SingleTaskGP


class OptimizerCont:
    def __init__(self, lb, ub, continuous_kern_type='smk', n_Cauchy=5, n_Gaussian=4, acq='ei', n_init=20):
        self.lb = lb
        self.ub = ub
        self.bounds = torch.stack((self.lb, self.ub))
        self.dim = len(self.lb)
        self.continuous_kern_type = continuous_kern_type
        self.n_Cauchy = n_Cauchy
        self.n_Gaussian = n_Gaussian
        self.acq = acq
        self.n_init = n_init
        self.X = None
        self.fX =None
        self.best_init_y = None

    def _create_initial_points(self):
        init_x = torch.rand(self.n_init, self.dim)
        init_x = self.bounds[0] + (self.bounds[1] - self.bounds[0]) * init_x
        return init_x

    def observe(self, x_next, y_next):
        self.X = torch.cat([self.X, x_next]) if self.X is not None else x_next
        self.fX = torch.cat([self.fX, y_next]) if self.fX is not None else y_next
        self.best_init_y = self.fX.min().item()

    def suggest(self, batch_size=1):
        if self.X is None:
            x_next = self._create_initial_points()
        else:
            x_next = self._get_next_points(batch_size=batch_size)
        return x_next


    def _get_next_points(self, batch_size):
        noise_prior = GammaPrior(1.1, 0.5)
        noise_prior_mode = (noise_prior.concentration - 1) / noise_prior.rate
        likelihood = GaussianLikelihood(
            noise_prior=noise_prior,
            batch_shape=[],
            noise_constraint=GreaterThan(
                # 0.000005,  # minimum observation noise assumed in the GP model
                0.0001,
                transform=None,
                initial_value=noise_prior_mode,
            ),
        )
        single_model = SingleTaskGP(self.X, self.fX, likelihood=likelihood, covar_module=self.continuous_kern_type,
                                    n_mixture=self.n_Gaussian+self.n_Gaussian, n_mixture1=self.n_Cauchy, n_mixture2=self.n_Gaussian)
        mll = ExactMarginalLogLikelihood(single_model.likelihood, single_model)
        fit_gpytorch_mll(mll)
        if self.acq == 'ei':
            acq_function = ExpectedImprovement(model=single_model, best_f=self.best_init_y, maximize=False)
        elif self.acq == 'pi':
            acq_function = ProbabilityOfImprovement(model=single_model, best_f=self.best_init_y, maximize=False)
        elif self.acq == 'ucb':
            acq_function = UpperConfidenceBound(model=single_model, beta=0.5, maximize=False)
        else:
            raise ValueError('Acquisition function not identified')
        candidates, _ = optimize_acqf(acq_function=acq_function, bounds=self.bounds, q=batch_size, num_restarts=100,
                                      raw_samples=512, options={"batch_limit": 5, "maxiter": 100})
        return candidates


import math
import sys
from copy import deepcopy

import gpytorch
import numpy as np
import torch

from SMKBO.cas.localbo_utils import train_gp
from SMKBO.cas.localbo_utils import random_sample_within_discrete_tr_ordinal


class CASMOPOLITANCat:
    """

    Parameters
    ----------
    f : function handle
    lb : Lower variable bounds, numpy.array, shape (d,).
    ub : Upper variable bounds, numpy.array, shape (d,).
    n_init : Number of initial points (2*dim is recommended), int.
    max_evals : Total evaluation budget, int.
    batch_size : Number of points in each batch, int.
    verbose : If you want to print information about the optimization progress, bool.
    use_ard : If you want to use ARD for the GP kernel.
    max_cholesky_size : Largest number of training points where we use Cholesky, int
    n_training_steps : Number of training steps for learning the GP hypers, int
    min_cuda : We use float64 on the CPU if we have this or fewer datapoints
    device : Device to use for GP fitting ("cpu" or "cuda")
    dtype : Dtype to use for GP fitting ("float32" or "float64")

    Data types that require special treatments
    cat_dims: list of lists. e.g. [[1, 2], [3, 4, 5]], which denotes that indices 1,2,3,4,5 are categorical, and [1, 2]
        belong to the same variable (a categorical variable with 2 possible values) and [3, 4, 5] belong to another,
        with 3 possible values.
    int_dims: list. [2, 3, 4]. Denotes the indices of the dimensions that are of integer types

    true_dim: The actual dimension of the problem. When there is no categorical variables, this value would be the same
        as the dimensionality inferred from the data. When there are categorical variable(s), due to the one-hot
        transformation. If not supplied, the dimension inferred from the data will be used.

    """

    def __init__(
            self,
            dim,
            n_init,
            max_evals,
            config,
            batch_size=1,
            verbose=True,
            use_ard=True,
            max_cholesky_size=2000,
            n_training_steps=50,
            min_cuda=1024,
            device="cpu",
            dtype="float32",
            acq='thompson',
            kernel_type='transformed_overlap',
            **kwargs
    ):

        # Very basic input checks
        # assert lb.ndim == 1 and ub.ndim == 1
        # assert len(lb) == len(ub)
        # assert np.all(ub > lb)
        assert max_evals > 0 and isinstance(max_evals, int)
        assert n_init > 0 and isinstance(n_init, int)
        assert batch_size > 0 and isinstance(batch_size, int)
        assert isinstance(verbose, bool) and isinstance(use_ard, bool)
        assert max_cholesky_size >= 0 and isinstance(batch_size, int)
        assert n_training_steps >= 30 and isinstance(n_training_steps, int)
        assert max_evals > n_init and max_evals > batch_size
        assert device == "cpu" or device == "cuda"
        assert dtype == "float32" or dtype == "float64"
        if device == "cuda":
            assert torch.cuda.is_available(), "can't use cuda if it's not available"

        # Save function information
        self.dim = dim
        self.config = config
        self.kwargs = kwargs
        # self.lb = lb
        # self.ub = ub

        # Settings
        self.n_init = n_init
        self.max_evals = max_evals
        self.batch_size = batch_size
        self.verbose = verbose
        self.use_ard = use_ard
        self.max_cholesky_size = max_cholesky_size
        self.n_training_steps = n_training_steps

        self.acq = acq
        self.kernel_type = kernel_type

        # Hyperparameters
        self.mean = np.zeros((0, 1))
        self.signal_var = np.zeros((0, 1))
        self.noise_var = np.zeros((0, 1))
        self.lengthscales = np.zeros((0, self.dim)) if self.use_ard else np.zeros((0, 1))

        # Tolerances and counters
        self.n_cand = kwargs['n_cand'] if 'n_cand' in kwargs.keys() else min(100 * self.dim, 5000)
        self.tr_multiplier = kwargs['multiplier'] if 'multiplier' in kwargs.keys() else 1.5
        self.failtol = kwargs['failtol'] if 'failtol' in kwargs.keys() else 40
        self.succtol = kwargs['succtol'] if 'succtol' in kwargs.keys() else 3
        self.n_evals = 0

        # Trust region sizes
        self.length_min = kwargs['length_min'] if 'length_min' in kwargs.keys() else 0.5 ** 7
        self.length_max = kwargs['length_max'] if 'length_max' in kwargs.keys() else 1.6
        self.length_init = kwargs['length_init'] if 'length_init' in kwargs.keys() else 0.8

        # Trust region sizes (in terms of Hamming distance) of the discrete variables.
        self.length_min_discrete = kwargs['length_min_discrete'] if 'length_min_discrete' in kwargs.keys() else 1
        self.length_max_discrete = kwargs['length_max_discrete'] if 'length_max_discrete' in kwargs.keys() else 30
        self.length_init_discrete = kwargs['length_init_discrete'] if 'length_init_discrete' in kwargs.keys() else 20

        # Save the full history
        self.X = np.zeros((0, self.dim))
        self.fX = np.zeros((0, 1))

        # Device and dtype for GPyTorch
        self.min_cuda = min_cuda
        self.dtype = torch.float32 if dtype == "float32" else torch.float64
        self.device = torch.device("cuda") if device == "cuda" else torch.device("cpu")
        if self.verbose:
            print("Using dtype = %s \nUsing device = %s" % (self.dtype, self.device))
            sys.stdout.flush()

        self._restart()

    def _restart(self):
        self._X = []
        self._fX = []
        self.failcount = 0
        self.succcount = 0
        self.length = self.length_init
        self.length_discrete = self.length_init_discrete

    def _adjust_length(self, fX_next):
        if np.min(fX_next) <= np.min(self._fX) - 1e-3 * math.fabs(np.min(self._fX)):
            self.succcount += 1
            self.failcount = 0
        else:
            self.succcount = 0
            self.failcount += 1

        if self.succcount == self.succtol:  # Expand trust region
            self.length = min([self.tr_multiplier * self.length, self.length_max])
            # For the Hamming distance-bounded trust region, we additively (instead of multiplicatively) adjust.
            self.length_discrete = int(min(self.length_discrete * self.tr_multiplier, self.length_max_discrete))
            # self.length = min(self.length * 1.5, self.length_max)
            self.succcount = 0
            print("expand", self.length, self.length_discrete)
        elif self.failcount == self.failtol:  # Shrink trust region
            # self.length = max([self.length_min, self.length / 2.0])
            self.failcount = 0
            # Ditto for shrinking.
            self.length_discrete = int(self.length_discrete / self.tr_multiplier)
            # self.length = max(self.length / 1.5, self.length_min)
            print("Shrink", self.length, self.length_discrete)

    def _create_and_select_candidates(self, X, fX, length, n_training_steps, hypers, return_acq=False):
        # assert X.min() >= 0.0 and X.max() <= 1.0
        # Figure out what device we are running on
        if len(X) < self.min_cuda:
            device, dtype = torch.device("cpu"), torch.float32
        else:
            device, dtype = self.device, self.dtype
        with gpytorch.settings.max_cholesky_size(self.max_cholesky_size):
            X_torch = torch.tensor(X).to(device=device, dtype=dtype)
            y_torch = torch.tensor(fX).to(device=device, dtype=dtype)
            gp = train_gp(
                train_x=X_torch, train_y=y_torch, use_ard=self.use_ard, num_steps=n_training_steps, hypers=hypers,
                kern=self.kernel_type,
                noise_variance=self.kwargs['noise_variance'] if
                'noise_variance' in self.kwargs else None
            )
            # Save state dict
            hypers = gp.state_dict()
        # Standardize function values.
        # mu, sigma = np.median(fX), fX.std()
        # sigma = 1.0 if sigma < 1e-6 else sigma
        # fX = (deepcopy(fX) - mu) / sigma

        from .localbo_utils import local_search
        x_center = X[fX.argmin().item(), :][None, :]

        def thompson(n_cand=5000):
            """Thompson sampling"""
            # Generate n_cand of candidates
            X_cand = np.array([
                random_sample_within_discrete_tr_ordinal(x_center[0], length, self.config)
                for _ in range(n_cand)
            ])
            with torch.no_grad(), gpytorch.settings.max_cholesky_size(self.max_cholesky_size):
                X_cand_torch = torch.tensor(X_cand, dtype=torch.float32)
                y_cand = gp.likelihood(gp(X_cand_torch)).sample(torch.Size([self.batch_size])).t().cpu().detach().numpy()
            # Revert the normalization process
            # y_cand = mu + sigma * y_cand

            # Select the best candidates
            X_next = np.ones((self.batch_size, self.dim))
            y_next = np.ones((self.batch_size, 1))
            for i in range(self.batch_size):
                indbest = np.argmin(y_cand[:, i])
                X_next[i, :] = deepcopy(X_cand[indbest, :])
                y_next[i, :] = deepcopy(y_cand[indbest, i])
                y_cand[indbest, :] = np.inf
            return X_next, y_next

        def _ei(X, augmented=True):
            """Expected improvement (with option to enable augmented EI"""
            from torch.distributions import Normal
            if not isinstance(X, torch.Tensor):
                X = torch.tensor(X, dtype=torch.float32)
            if X.dim() == 1:
                X = X.reshape(1, -1)
            gauss = Normal(torch.zeros(1), torch.ones(1))
            # flip for minimization problems
            preds = gp(X)
            mean, std = -preds.mean, preds.stddev
            # use in-fill criterion
            mu_star = -gp.likelihood(gp(torch.tensor(x_center[0].reshape(1, -1), dtype=torch.float32))).mean

            u = (mean - mu_star) / std
            ucdf = gauss.cdf(u)
            updf = torch.exp(gauss.log_prob(u))
            ei = std * updf + (mean - mu_star) * ucdf
            if augmented:
                sigma_n = gp.likelihood.noise
                ei *= (1. - torch.sqrt(torch.tensor(sigma_n)) / torch.sqrt(sigma_n + std ** 2))
            return ei

        def _ucb(X, beta=5.):
            """Upper confidence bound"""
            if not isinstance(X, torch.Tensor):
                X = torch.tensor(X, dtype=torch.float32)
            if X.dim() == 1:
                X = X.reshape(1, -1)
            # Invoked when you supply X in one-hot representations

            preds = gp.likelihood(gp(X))
            mean, std = preds.mean, preds.stddev
            return -(mean + beta * std)

        if self.acq in ['ei', 'ucb']:
            if self.batch_size == 1:
                # Sequential setting
                if self.acq == 'ei':
                    X_next, acq_next = local_search(x_center[0], _ei, self.config, length, 3, self.batch_size)
                else:
                    X_next, acq_next = local_search(x_center[0], _ucb, self.config, length, 3, self.batch_size)

            else:
                # batch setting: for these, we use the fantasised points {x, y}
                X_next = torch.tensor([], dtype=torch.float32)
                acq_next = np.array([])
                for p in range(self.batch_size):
                    if self.acq == 'ei':
                        x_next, acq = local_search(x_center[0], _ei, self.config, length, 3, 1)
                    else:
                        x_next, acq = local_search(x_center[0], _ucb, self.config, length, 3, 1)
                    x_next = torch.tensor(x_next, dtype=torch.float32)
                    # The fantasy point is filled by the posterior mean of the Gaussian process.
                    y_next = gp(x_next).mean.detach()
                    with gpytorch.settings.max_cholesky_size(self.max_cholesky_size):
                        X_torch = torch.cat((X_torch, x_next), dim=0)
                        y_torch = torch.cat((y_torch, y_next), dim=0)
                        gp = train_gp(
                            train_x=X_torch, train_y=y_torch, use_ard=self.use_ard, num_steps=n_training_steps,
                            kern=self.kernel_type,
                            hypers=hypers,
                            noise_variance=self.kwargs['noise_variance'] if
                            'noise_variance' in self.kwargs else None
                        )
                    X_next = torch.cat((X_next, x_next), dim=0)
                    acq_next = np.hstack((acq_next, acq))

        elif self.acq == 'thompson':
            X_next, acq_next = thompson()
        else:
            raise ValueError('Unknown acquisition function choice %s' % self.acq)

        # Remove the torch tensors
        del X_torch, y_torch
        X_next = np.array(X_next)
        if return_acq:
            return X_next, acq_next
        return X_next
